# Copyright 2021-2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
import abc
import contextlib
import json
import logging
import sys
import time
import types
import zlib
import collections

from kazoo.exceptions import (
    LockTimeout,
    NoNodeError,
    NodeExistsError,
    NotEmptyError,
)
from kazoo.retry import KazooRetry

from zuul.zk import sharding
from zuul.zk import ZooKeeperClient
from zuul.zk.exceptions import LockException
from zuul.zk.locks import SessionAwareLock


class BaseZKContext:
    profile_logger = logging.getLogger('zuul.profile')
    profile_default = False
    default_lock_identifier = None
    # Only changed by unit tests.
    # The default scales with number of procs.
    _max_workers = None

    def __init__(self):
        # We create the executor dict in enter to make sure that this
        # is used as a context manager and cleaned up properly.
        self.executor = None

    def __enter__(self):
        if self.executor:
            raise RuntimeError("ZKContext entered multiple times")
        # This is a dictionary keyed by class.  ZKObject subclasses
        # can request a dedicated ThreadPoolExecutor for their class
        # so that deserialize methods that use it can avoid deadlocks
        # with child class deserialize methods.
        self.executor = collections.defaultdict(
            lambda: ThreadPoolExecutor(
                max_workers=self._max_workers,
                thread_name_prefix="ZKContext",
            ))
        return self

    def __exit__(self, etype, value, tb):
        if self.executor:
            for executor in self.executor.values():
                if sys.version_info >= (3, 9):
                    executor.shutdown(wait=False, cancel_futures=True)
                else:
                    executor.shutdown(wait=False)
        self.executor = None


class ZKContext(BaseZKContext):
    def __init__(self, zk_client, lock, stop_event, log,
                 default_lock_identifier=None):
        super().__init__()
        if isinstance(zk_client, ZooKeeperClient):
            client = zk_client.client
        else:
            client = zk_client
        self.client = client
        self.lock = lock
        self.stop_event = stop_event
        self.log = log
        self.cumulative_read_time = 0.0
        self.cumulative_write_time = 0.0
        self.cumulative_read_objects = 0
        self.cumulative_write_objects = 0
        self.cumulative_read_znodes = 0
        self.cumulative_write_znodes = 0
        self.cumulative_read_bytes = 0
        self.cumulative_write_bytes = 0
        self.build_references = False
        self.profile = self.profile_default
        self.default_lock_identifier = default_lock_identifier

    def sessionIsValid(self):
        return (not self.lock or self.lock.is_still_valid())

    def sessionIsInvalid(self):
        return not self.sessionIsValid()

    def shouldAbortRetry(self):
        return (self.sessionIsInvalid() or
                (self.stop_event and self.stop_event.is_set()))

    def updateStatsFromOtherContext(self, other):
        self.cumulative_read_time += other.cumulative_read_time
        self.cumulative_write_time += other.cumulative_write_time
        self.cumulative_read_objects += other.cumulative_read_objects
        self.cumulative_write_objects += other.cumulative_write_objects
        self.cumulative_read_znodes += other.cumulative_read_znodes
        self.cumulative_write_znodes += other.cumulative_write_znodes
        self.cumulative_read_bytes += other.cumulative_read_bytes
        self.cumulative_write_bytes += other.cumulative_write_bytes

    def profileEvent(self, etype, path):
        if not self.profile:
            return
        self.profile_logger.debug(
            'ZK 0x%x %s %s  '
            'rt=%s wt=%s  ro=%s wo=%s  rn=%s wn=%s  rb=%s wb=%s',
            id(self), etype, path,
            self.cumulative_read_time, self.cumulative_write_time,
            self.cumulative_read_objects, self.cumulative_write_objects,
            self.cumulative_read_znodes, self.cumulative_write_znodes,
            self.cumulative_read_bytes, self.cumulative_write_bytes)


class LocalZKContext(BaseZKContext):
    """A Local ZKContext that means don't actually write anything to ZK"""

    def __init__(self, log):
        super().__init__()
        self.client = None
        self.lock = None
        self.stop_event = None
        self.log = log

    def sessionIsValid(self):
        return True

    def sessionIsInvalid(self):
        return False


class ZKObject:
    _retry_interval = 5
    _zkobject_compressed_size = 0
    _zkobject_uncompressed_size = 0
    io_reader_class = sharding.RawZKIO
    io_writer_class = sharding.RawZKIO
    truncate_on_create = False
    delete_on_error = False
    log_error_missing = True
    makepath = True

    # Implementations of these two methods are required
    def getPath(self):
        """Return the path to save this object in ZK

        :returns: A string representation of the Znode path
        """
        raise NotImplementedError()

    def serialize(self, context):
        """Implement this method to return the data to save in ZK.

        :returns: A byte string
        """
        raise NotImplementedError()

    # This should work for most classes
    def deserialize(self, data, context, extra=None):
        """Implement this method to convert serialized data into object
        attributes.

        :param bytes data: A byte string to deserialize
        :param ZKContext context: A ZKContext object with the current
            ZK session and lock.
        :param extra dict: A dictionary of extra data for use in
            deserialization.

        :returns: A dictionary of attributes and values to be set on
        the object.

        """
        if isinstance(data, dict):
            return data
        return json.loads(data.decode('utf-8'))

    # These methods are public and shouldn't need to be overridden
    def updateAttributes(self, context, **kw):
        """Update attributes on this object and save to ZooKeeper

        Instead of using attribute assignment, call this method to
        update attributes on this object.  It will update the local
        values and also write out the updated object to ZooKeeper.

        :param ZKContext context: A ZKContext object with the current
            ZK session and lock.  Be sure to acquire the lock before
            calling methods on this object.  This object will validate
            that the lock is still valid before writing to ZooKeeper.

        All other parameters are keyword arguments which are
        attributes to be set.  Set as many attributes in one method
        call as possible for efficient network use.
        """
        old = self.__dict__.copy()
        self._set(**kw)
        serial = self._trySerialize(context)
        if hash(serial) != getattr(self, '_zkobject_hash', None):
            try:
                self._save(context, serial)
            except Exception:
                # Roll back our old values if we aren't able to update ZK.
                self._set(**old)
                raise

    @contextlib.contextmanager
    def activeContext(self, context):
        if self._active_context:
            raise RuntimeError(
                f"Another context is already active {self._active_context}")
        try:
            old = self.__dict__.copy()
            self._set(_active_context=context)
            yield
            serial = self._trySerialize(context)
            if hash(serial) != getattr(self, '_zkobject_hash', None):
                try:
                    self._save(context, serial)
                except Exception:
                    # Roll back our old values if we aren't able to update ZK.
                    self._set(**old)
                    raise
        finally:
            self._set(_active_context=None)

    @classmethod
    def new(klass, context, **kw):
        """Create a new instance and save it in ZooKeeper"""
        obj = klass()
        obj._set(**kw)
        data = obj._trySerialize(context)
        obj._save(context, data, create=True)
        return obj

    @classmethod
    def fromZK(klass, context, path, **kw):
        """Instantiate a new object from data in ZK"""
        obj = klass()
        obj._set(**kw)
        obj._load(context, path=path)
        return obj

    def internalCreate(self, context):
        """Create the object in ZK from an existing ZKObject

        This should only be used in special circumstances: when we
        know it's safe to start using a ZKObject before it's actually
        created in ZK.  Normally use .new()
        """
        data = self._trySerialize(context)
        self._save(context, data, create=True)

    def refresh(self, context):

        """Update data from ZK"""
        self._load(context)

    def exists(self, context):
        """Return whether the object exists in ZK"""
        path = self.getPath()
        return bool(context.client.exists(path))

    def _trySerialize(self, context):
        if isinstance(context, LocalZKContext):
            return b''
        try:
            return self.serialize(context)
        except Exception:
            # A higher level must handle this exception, but log
            # ourself here so we know what object triggered it.
            context.log.error(
                "Exception serializing ZKObject %s", self)
            raise

    @classmethod
    def _delete(cls, context, path):
        if context.sessionIsInvalid():
            raise Exception("ZooKeeper session or lock not valid")
        cls._retry(context, context.client.delete,
                   path, recursive=True)
        context.profileEvent('delete', path)

    def delete(self, context):
        path = self.getPath()
        try:
            self._delete(context, path)
        except Exception:
            context.log.error(
                "Exception deleting ZKObject %s at %s", self, path)
            raise

    def estimateDataSize(self, seen=None):
        """Attempt to find all ZKObjects below this one and sum their
        compressed and uncompressed sizes.

        :returns: (compressed_size, uncompressed_size)
        """
        compressed_size = self._zkobject_compressed_size
        uncompressed_size = self._zkobject_uncompressed_size

        if seen is None:
            seen = {self}

        def walk(obj):
            compressed = 0
            uncompressed = 0
            if isinstance(obj, ZKObject):
                if obj in seen:
                    return 0, 0
                seen.add(obj)
                compressed, uncompressed = obj.estimateDataSize(seen)
            elif (isinstance(obj, dict) or
                  isinstance(obj, types.MappingProxyType)):
                for sub in list(obj.values()):
                    c, u = walk(sub)
                    compressed += c
                    uncompressed += u
            elif (isinstance(obj, list) or
                  isinstance(obj, tuple)):
                for sub in obj:
                    c, u = walk(sub)
                    compressed += c
                    uncompressed += u
            return compressed, uncompressed

        c, u = walk(self.__dict__)
        compressed_size += c
        uncompressed_size += u

        return (compressed_size, uncompressed_size)

    def getZKVersion(self):
        """Return the ZK version of the object as of the last load/refresh.

        Returns None if the object is newly created.
        """
        zstat = getattr(self, '_zstat', None)
        # If zstat is None, we created the object
        if zstat is None:
            return None
        return zstat.version

    def getZKczxid(self):
        """Return the ZK create transaction id of the object.

        Returns None if the object is newly created.
        """
        zstat = getattr(self, '_zstat', None)
        # If zstat is None, we created the object
        if zstat is None:
            return None
        return zstat.czxid

    def getZKmzxid(self):
        """Return the ZK last modified transaction id of the object as
        of the last load/refresh.

        Returns None if the object is newly created.

        """
        zstat = getattr(self, '_zstat', None)
        # If zstat is None, we created the object
        if zstat is None:
            return None
        return zstat.mzxid

    # Private methods below

    @classmethod
    def _retry(cls, context, func, *args, max_tries=-1, **kw):
        kazoo_retry = KazooRetry(max_tries=max_tries,
                                 interrupt=context.shouldAbortRetry,
                                 delay=cls._retry_interval, backoff=0,
                                 ignore_expire=False)
        try:
            return kazoo_retry(func, *args, **kw)
        except InterruptedError:
            pass

    def __init__(self):
        # Don't support any arguments in constructor to force us to go
        # through a save or restore path.
        super().__init__()
        self._set(_active_context=None)

    @staticmethod
    def _retryableLoad(io_class, context, path):
        with io_class(context.client, path) as stream:
            compressed_data = stream.read()
            zstat = stream.zstat
        context.cumulative_read_time += stream.cumulative_read_time
        context.cumulative_read_objects += 1
        context.cumulative_read_znodes += stream.znodes_read
        context.cumulative_read_bytes += stream.bytes_read
        return compressed_data, zstat

    def _load(self, context, path=None):
        if path is None:
            path = self.getPath()
        compressed_data, zstat = self._loadData(context, path)
        self._updateFromRaw(context, compressed_data, zstat, None)

    @classmethod
    def _loadData(cls, context, path):
        if context.sessionIsInvalid():
            raise Exception("ZooKeeper session or lock not valid")
        try:
            compressed_data, zstat = cls._retry(context, cls._retryableLoad,
                                                cls.io_reader_class,
                                                context, path)
            context.profileEvent('get', path)
        except Exception as exc:
            if cls.log_error_missing or not isinstance(exc, NoNodeError):
                context.log.error(
                    "Exception loading ZKObject at %s", path)
            if cls.delete_on_error:
                cls._delete(context, path)
            raise
        return compressed_data, zstat

    @classmethod
    def _fromRaw(cls, context, raw_data, zstat, extra, **kw):
        obj = cls()
        obj._updateFromRaw(context, raw_data, zstat, extra)
        return obj

    def _updateFromRaw(self, context, raw_data, zstat, extra):
        try:
            self._set(_zkobject_hash=None)
            data = self._decompressData(raw_data)
            self._set(**self.deserialize(data, context, extra))
            self._set(_zstat=zstat,
                      _zkobject_hash=hash(data),
                      _zkobject_compressed_size=len(raw_data),
                      _zkobject_uncompressed_size=len(data))
        except Exception:
            if self.delete_on_error:
                self.delete(context)
            raise

    @classmethod
    def _compressData(cls, data):
        return zlib.compress(data)

    @classmethod
    def _decompressData(cls, raw_data):
        try:
            return zlib.decompress(raw_data)
        except zlib.error:
            # Fallback for old, uncompressed data
            return raw_data

    @staticmethod
    def _retryableSave(io_class, context, create, makepath, path, data,
                       version):
        zstat = None
        with io_class(context.client, path, create=create, makepath=makepath,
                      version=version) as stream:
            stream.truncate(0)
            stream.write(data)
            stream.flush()
            context.cumulative_write_time += stream.cumulative_write_time
            context.cumulative_write_objects += 1
            context.cumulative_write_znodes += stream.znodes_written
            context.cumulative_write_bytes += stream.bytes_written
            zstat = stream.zstat
        return zstat

    def _save(self, context, data, create=False):
        if isinstance(context, LocalZKContext):
            return
        path = self.getPath()
        if context.sessionIsInvalid():
            raise Exception("ZooKeeper session or lock not valid")
        compressed_data = self._compressData(data)

        try:
            if create and not self.truncate_on_create:
                exists = self._retry(context, context.client.exists, path)
                context.profileEvent('exists', path)
                if exists is not None:
                    raise NodeExistsError
            zstat = getattr(self, '_zstat', None)
            if zstat is not None:
                version = self._zstat.version
            else:
                version = -1
            zstat = self._retry(context, self._retryableSave,
                                self.io_writer_class, context, create,
                                self.makepath, path, compressed_data, version)
            context.profileEvent('set', path)
        except Exception:
            context.log.error(
                "Exception saving ZKObject %s at %s", self, path)
            raise
        self._set(_zstat=zstat,
                  _zkobject_hash=hash(data),
                  _zkobject_compressed_size=len(compressed_data),
                  _zkobject_uncompressed_size=len(data),
                  )

    def __setattr__(self, name, value):
        if self._active_context:
            super().__setattr__(name, value)
        else:
            raise Exception("Unable to modify ZKObject %s" %
                            (repr(self),))

    def _set(self, **kw):
        for name, value in kw.items():
            super().__setattr__(name, value)


class ShardedZKObject(ZKObject):
    # If the node exists when we create we normally error, unless this
    # is set, in which case we proceed and truncate.
    truncate_on_create = False
    # Normally we delete nodes which have syntax errors, but the
    # pipeline summary is read without a write lock, so those are
    # expected.  Don't delete them in that case.
    delete_on_error = True
    io_reader_class = sharding.BufferedShardReader
    io_writer_class = sharding.BufferedShardWriter


class LockableZKObject(ZKObject):
    _lock = None
    _deleted = False

    def __new__(klass, *args, **kwargs):
        zko = super().__new__(klass)
        zko._set(_lock_contenders=dict())
        return zko

    def getLockPath(self):
        """Return the path for the lock of this object in ZK

        :returns: A string representation of the Znode path
        """
        raise NotImplementedError()

    @classmethod
    def new(klass, context, **kw):
        """Create a new instance and save it in ZooKeeper"""
        obj = klass()
        obj._set(**kw)
        # Create the lock path first.  In the future, if we want to
        # support creating locked objects, we can acquire it here.
        obj._createLockPath(context)
        data = obj._trySerialize(context)
        obj._save(context, data, create=True)
        return obj

    def internalCreate(self, context):
        self._createLockPath(context)
        super().internalCreate(context)

    def _createLockPath(self, context):
        if isinstance(context, LocalZKContext):
            return
        path = self.getLockPath()
        if context.sessionIsInvalid():
            raise Exception("ZooKeeper session or lock not valid")
        try:
            self._retry(context, context.client.ensure_path, path)
        except Exception:
            context.log.error(
                "Exception creating ZKObject %s lock path at %s", self, path)
            raise

    def _deleteLockPath(self, client):
        path = self.getLockPath()
        # Give other actors 30 seconds to realize this object
        # doesn't exist anymore and they should stop trying to
        # lock it.
        for x in range(30):
            # This handles connection-related retries, but not
            # conflicts.  We ignore the session here because we're
            # trying to delete a lock; we want to do that even if we
            # lose whatever lock the context might have.
            kazoo_retry = KazooRetry(max_tries=-1,
                                     delay=self._retry_interval,
                                     backoff=0)
            try:
                return kazoo_retry(client.delete, path, recursive=True)
            except NotEmptyError:
                time.sleep(1)

    @contextmanager
    def locked(self, context, blocking=True, timeout=None):
        if not (lock := self.acquireLock(context, blocking=blocking,
                                         timeout=timeout)):
            raise LockException(f"Failed to acquire lock on {self}")
        try:
            yield lock
        finally:
            try:
                self.releaseLock(context)
            except Exception:
                context.log.exception("Failed to release lock on %s", self)

    def acquireLock(self, context, blocking=True, timeout=None,
                    identifier=None):
        have_lock = False
        lock = None
        path = self.getLockPath()
        identifier = identifier or context.default_lock_identifier
        try:
            # We create the lock path when we create the object in ZK,
            # so there is no need to ensure the path on lock.  This
            # lets us avoid re-creating the lock if the object was
            # deleted behind our back.
            lock = SessionAwareLock(context.client, path,
                                    identifier=identifier,
                                    ensure_path=False)
            have_lock = lock.acquire(blocking, timeout)
        except NoNodeError:
            # Request disappeared
            have_lock = False
        except LockTimeout:
            have_lock = False
            context.log.error("Timeout trying to acquire lock: %s", path)

        # If we aren't blocking, it's possible we didn't get the lock
        # because someone else has it.
        if not have_lock:
            return None

        self._set(_lock=lock)
        return lock

    def releaseLock(self, context):
        if self._lock is None:
            return
        self._lock.release()
        self._set(_lock=None)
        if self._deleted:
            # If we are releasing the lock after deleting the object,
            # also cleanup the lock path.
            try:
                self._deleteLockPath(context.client)
            except Exception:
                context.log.error("Unable to delete lock path for %s", self)

    def delete(self, context):
        super().delete(context)
        self._set(_deleted=True)
        # Most lockable zkobjects are deleted while holding the lock,
        # so we let releaseLock delete the lock path.  But for those
        # that aren't, we should delete the lock path now.
        if not self.hasLock():
            try:
                self._deleteLockPath(context.client)
            except Exception:
                context.log.error("Unable to delete lock path for %s", self)

    def hasLock(self):
        if self._lock is None:
            return False
        return self._lock.is_still_valid()


class PolymorphicZKObjectMixin(abc.ABC):

    # Make this mixin a true abstract base class that can't be
    # instantiated. The '_subclass_id' property is automatically
    # "implemented" through subclassing (see '__init_subclass__').
    @property
    @abc.abstractmethod
    def _subclass_id(self):
        pass

    # Mapping of subclass id to subclass for a particular abstract
    # parent class. Creating an instance without a subclass id
    # starts a new hierarchy. This is only supported when none
    # of the parent classes is alread a variable ZKObject class.
    @property
    @abc.abstractmethod
    def _subclasses(self):
        pass

    def __init_subclass__(cls, *, subclass_id=None, **kwargs):
        super().__init_subclass__(**kwargs)
        if subclass_id is None:
            if isinstance(cls._subclasses, dict):
                raise TypeError(
                    "Can't create new hierarchy below existing variable base")
            # This is a new "base" class
            cls._subclasses = {}
        else:
            if not isinstance(cls._subclasses, dict):
                raise TypeError(
                    "Can't create a subclass without variable base class")
            # "Implement" the abstract '_subclass_id' property
            sid = subclass_id.encode("utf8")
            if sid in cls._subclasses:
                raise ValueError(
                    f"Subclass with id {subclass_id} already exists")
            cls._subclass_id = sid
            cls._subclasses[sid] = cls

    @classmethod
    def fromZK(cls, context, path, **kw):
        raw_data, zstat = cls._loadData(context, path)
        return cls._fromRaw(context, raw_data, zstat, None, **kw)

    @classmethod
    def _compressData(cls, data):
        compressed_data = super()._compressData(data)
        return b"\0".join((cls._subclass_id, compressed_data))

    @classmethod
    def _decompressData(cls, raw_data):
        subclass_id, _, compressed_data = raw_data.partition(b"\0")
        return super()._decompressData(compressed_data)

    @classmethod
    def _fromRaw(cls, context, raw_data, zstat, extra, **kw):
        subclass_id, _, _ = raw_data.partition(b"\0")
        try:
            klass = cls._subclasses[subclass_id]
        except KeyError:
            raise RuntimeError(f"Unknown subclass id: {subclass_id}")
        return super(
            PolymorphicZKObjectMixin, klass)._fromRaw(
                context, raw_data, zstat, extra, **kw)



# Copyright 2017 Red Hat, Inc.
# Copyright 2024-2025 Acme Gating, LLC
#
# Zuul is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Zuul is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# This is not needed in python3 - but it is needed in python2 because there
# is a json module in ansible.plugins.callback and python2 gets confused.
# Easy local testing with ansible-playbook is handy when hacking on zuul_stream
# so just put in the __future__ statement.
from __future__ import absolute_import

DOCUMENTATION = '''
    callback: zuul_stream
    short_description: This is the Zuul streaming callback
    version_added: "2.4"
    description:
        - This callback is necessary for Zuul to properly stream logs from the
          remote test node.
    type: stdout
    extends_documentation_fragment:
      - default_callback
    requirements:
      - Set as stdout in config
'''

import datetime
import logging
import logging.config
import json
import os
import re2
import socket
import threading
import time

from ansible.playbook.block import Block
from ansible.plugins.callback import default
from ansible.module_utils._text import to_text
from ansible.module_utils.parsing.convert_bool import boolean
from zuul.ansible import paths

from zuul.ansible import logconfig

LOG_STREAM_VERSION = 0

DEFAULT_POSIX_PORT = 19885
DEFAULT_WINDOWS_PORT = 19886

POSIX_ACTIONS = (
    'command',
    'shell',
    'ansible.builtin.command',
    'ansible.builtin.shell',
)

WINDOWS_ACTIONS = (
    'win_command',
    'win_shell',
    'ansible.windows.win_command',
    'ansible.windows.win_shell',
)

ACTION_LOG_STREAM_PORT = {
    **{a: DEFAULT_POSIX_PORT for a in POSIX_ACTIONS},
    **{a: DEFAULT_WINDOWS_PORT for a in WINDOWS_ACTIONS},
}

# Actions that support live log streaming
STREAMING_ACTIONS = POSIX_ACTIONS + WINDOWS_ACTIONS

# Actions that produce stdout/err that should go in the log from JSON
OUTPUT_ACTIONS = ('raw',)

# Actions that produce stdout/err that should go in the log (whether
# from streaming or JSON).
ALL_ACTIONS = STREAMING_ACTIONS + OUTPUT_ACTIONS


def zuul_filter_result(result):
    """Remove keys from shell/command output.

    Zuul streams stdout/stderr into the log above, so including stdout and
    stderr in the result dict that ansible displays in the logs is duplicate
    noise. We keep stdout/stderr in the result dict so that other callback
    plugins like ARA could also have access to it. But drop them here.

    Remove changed so that we don't show a bunch of "changed" titles
    on successful shell tasks, since that doesn't make sense from a Zuul
    POV. The super class treats missing "changed" key as False.

    Remove cmd because most of the script content where people want to
    see the script run is run with -x. It's possible we may want to revist
    this to be smarter about when we remove it - like, only remove it
    if it has an embedded newline - so that for normal 'simple' uses
    of cmd it'll echo what the command was for folks.
    """

    stdout = result.pop('stdout', '')
    stdout_lines = result.pop('stdout_lines', [])
    if not stdout_lines and stdout:
        stdout_lines = stdout.split('\n')

    stderr = result.pop('stderr', '')
    stderr_lines = result.pop('stderr_lines', [])
    if not stderr_lines and stderr:
        stderr_lines = stderr.split('\n')

    for key in ('changed', 'cmd', 'zuul_log_id', 'invocation'):
        result.pop(key, None)

    # Combine stdout / stderr
    return stdout_lines + stderr_lines


def is_rescuable(task):
    if task._parent is None:
        return False
    if isinstance(task._parent, Block):
        if task._parent.rescue:
            return True
    return is_rescuable(task._parent)


class Streamer:
    def __init__(self, callback, host, ip, port, log_id):
        self.callback = callback
        self.host = host
        self.ip = ip
        self.port = port
        self.log_id = log_id
        self._zuul_console_version = 0
        self.stopped = False
        self.connected = False

    def start(self):
        self.thread = threading.Thread(target=self._read_log)
        self.thread.daemon = True
        self.thread.start()

    def stop(self, grace):
        self.stop_ts = time.monotonic()
        self.stop_grace = grace
        self.stopped = True

    def join(self, timeout):
        self.thread.join(timeout)

    def is_alive(self):
        return self.thread.is_alive()

    def _read_log_connect(self):
        logger_retries = 0
        # We don't check stopped here because if we ever succeed in
        # connecting, then we should always try again; that's because
        # we may need to reconnect in order to send another command to
        # delete the log file.
        while True:
            try:
                s = socket.create_connection((self.ip, self.port), 5)
                # Disable the socket timeout after we have successfully
                # connected to accomodate the fact that jobs may not be writing
                # logs continously. Without this we can easily trip the 5
                # second timeout.
                s.settimeout(None)
                # Some remote commands can run for some time without
                # producing output.  In case there are network
                # components that might drop an idle TCP connection,
                # enable keepalives so that we can hopefully maintain
                # the connection, or at the least, be notified if it
                # is terminated.  Ping every 30 seconds after 30
                # seconds of idle activity.  Allow 3 minutes of lost
                # pings before we fail.
                s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 30)
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 30)
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 9)
                self.connected = True
                return s
            except socket.timeout:
                if not self.stopped:
                    self.callback._log_streamline(
                        "localhost",
                        "Timeout exception waiting for the logger. "
                        "Please check connectivity to [%s:%s]"
                        % (self.ip, self.port),
                        stopped=True)
                return None
            except Exception:
                if self.stopped:
                    return None
                if logger_retries % 10 == 0:
                    self.callback._log("[%s] Waiting on logger" % self.host)
                logger_retries += 1
                time.sleep(0.1)
                continue

    def _read_log(self):
        s = self._read_log_connect()
        if s is None:
            # Can't connect; _read_log_connect() already logged an
            # error for us, just bail
            return

        # If we were asked to stop during the connection attempt, and
        # the grace time has expired return now.  If we are stopped
        # for a skipped task, then there will be no grace time set, so
        # this should return immediately.
        if self.stopped:
            if time.monotonic() >= (self.stop_ts + self.stop_grace):
                return

        # Find out what version we are running against
        s.send(f'v:{LOG_STREAM_VERSION}\n'.encode('utf-8'))
        buff = s.recv(1024).decode('utf-8').strip()

        # NOTE(ianw) 2022-07-21 : zuul_console from < 6.3.0 do not
        # understand this protocol.  They will assume the send
        # above is a log request and send back the not found
        # message in a loop.  So to handle this we disconnect and
        # reconnect.  When we decide to remove this, we can remove
        # anything in the "version 0" path.
        if buff == '[Zuul] Log not found':
            s.shutdown(socket.SHUT_RDWR)
            s.close()
            s = self._read_log_connect()
            if s is None:
                return
        else:
            self._zuul_console_version = int(buff)

        if self._zuul_console_version >= 1:
            msg = 's:%s\n' % self.log_id
        else:
            msg = '%s\n' % self.log_id

        s.send(msg.encode("utf-8"))
        buff = s.recv(4096)
        buffering = True
        while buffering:
            if b'\n' in buff:
                (line, buff) = buff.split(b'\n', 1)

                stopped = False
                if self.stopped:
                    if time.monotonic() >= (self.stop_ts + self.stop_grace):
                        # The task is finished, but we're still
                        # reading the log.  We might just be reading
                        # more data, or we might be reading "Log not
                        # found" repeatedly.  That might be due to
                        # contention on a busy node and the log will
                        # show up soon, or it may be because we
                        # skipped the task and nothing ran.  In the
                        # latter case, the grace period will be 0, but
                        # in the former, we will allow an extra 10
                        # seconds for the log to show up.
                        stopped = True

                # We can potentially get binary data here. In order to
                # being able to handle that use the backslashreplace
                # error handling method. This decodes unknown utf-8
                # code points to escape sequences which exactly represent
                # the correct data without throwing a decoding exception.
                done = self.callback._log_streamline(
                    self.host, line.decode("utf-8", "backslashreplace"),
                    stopped=stopped)
                if done:
                    if self._zuul_console_version > 0:
                        try:
                            # reestablish connection and tell console to
                            # clean up
                            s.shutdown(socket.SHUT_RDWR)
                            s.close()
                            s = self._read_log_connect()
                            s.send(f'f:{self.log_id}\n'.encode('utf-8'))
                            s.shutdown(socket.SHUT_RDWR)
                            s.close()
                        except Exception:
                            # Don't worry if this fails
                            pass
                    return
            else:
                try:
                    more = s.recv(4096)
                except TimeoutError:
                    self.callback._log_streamline(
                        self.host,
                        "[Zuul] Lost log stream connection to [%s:%s]"
                        % (self.ip, self.port),
                        stopped=True)
                    raise
                if not more:
                    buffering = False
                else:
                    buff += more
        if buff:
            self.callback._log_streamline(
                self.host, buff.decode("utf-8", "backslashreplace"),
                stopped=True)


class CallbackModule(default.CallbackModule):

    '''
    This is the Zuul streaming callback. It's based on the default
    callback plugin, but streams results from shell commands.
    '''

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'stdout'
    CALLBACK_NAME = 'zuul_stream'

    def __init__(self):

        super(CallbackModule, self).__init__()
        self._task = None
        self._daemon_running = False
        self._play = None
        self._streamers = {}  # task uuid -> streamer
        self.sent_failure_result = False
        self.configure_logger()
        self.configure_regexes()
        self._items_done = False
        self._deferred_result = None
        self._playbook_name = None

    def configure_logger(self):
        # ansible appends timestamp, user and pid to the log lines emitted
        # to the log file. We're doing other things though, so we don't want
        # this.
        logging_config = logconfig.load_job_config(
            os.environ['ZUUL_JOB_LOG_CONFIG'])

        if self._display.verbosity > 2:
            logging_config.setDebug()

        logging_config.apply()

        self._logger = logging.getLogger('zuul.executor.ansible')
        self._result_logger = logging.getLogger(
            'zuul.executor.ansible.result')

    def configure_regexes(self):
        try:
            serialized = os.environ.get('ZUUL_JOB_FAILURE_OUTPUT', '[]')
            regexes = json.loads(serialized)
            self.failure_output = [re2.compile(x) for x in regexes]
        except Exception as e:
            self._log("Error compiling failure output regexes: %s"
                      % (str(e),), job=False, executor=True)
            self.failure_output = []

    def _log(self, msg, ts=None, job=True, executor=False, debug=False):
        # With the default "linear" strategy (and likely others),
        # Ansible will send the on_task_start callback, and then fork
        # a worker process to execute that task.  Since we spawn a
        # thread in the on_task_start callback, we can end up emitting
        # a log message in this method while Ansible is forking.  If a
        # forked process inherits a Python file object (i.e., stdout)
        # that is locked by a thread that doesn't exist in the fork
        # (i.e., this one), it can deadlock when trying to flush the
        # file object.  To minimize the chances of that happening, we
        # should avoid using _display outside the main thread.
        # Therefore:

        # Do not set executor=True from any calls from a thread
        # spawned in this callback.

        msg = msg.rstrip()
        if job:
            now = ts or datetime.datetime.now()
            self._logger.info("{now} | {msg}".format(now=now, msg=msg))
        if executor:
            if debug:
                self._display.vvv(msg)
            else:
                self._display.display(msg)

    def _log_streamline(self, host, line, stopped):
        if "[Zuul] Task exit code" in line:
            return True
        elif "[Zuul] Log not found" in line:
            # don't output this line
            if stopped:
                return True
            else:
                return False
        elif " | " in line:
            ts, ln = line.split(' | ', 1)

            self._log("%s | %s " % (host, ln), ts=ts)
            self._check_failure_output(host, ln)
            return False
        else:
            self._log("%s | %s " % (host, line))
            self._check_failure_output(host, line)
            return False

    def _check_failure_output(self, host, line):
        if self.sent_failure_result:
            return
        for fail_re in self.failure_output:
            if fail_re.match(line):
                self._log(
                    '%s | Early failure in job, matched regex "%s"' % (
                        host, fail_re.pattern,))
                self._result_logger.info("failure")
                self.sent_failure_result = True

    def _log_module_failure(self, result, result_dict):
        if 'module_stdout' in result_dict and result_dict['module_stdout']:
            self._log_message(
                result, status='MODULE FAILURE',
                msg=result_dict['module_stdout'])
        elif 'exception' in result_dict and result_dict['exception']:
            self._log_message(
                result, status='MODULE FAILURE',
                msg=result_dict['exception'])
        elif 'module_stderr' in result_dict:
            self._log_message(
                result, status='MODULE FAILURE',
                msg=result_dict['module_stderr'])

    def v2_playbook_on_start(self, playbook):
        self._playbook_name = os.path.splitext(playbook._file_name)[0]

    def v2_playbook_on_include(self, included_file):
        for host in included_file._hosts:
            self._log("{host} | included: {filename}".format(
                host=host.name,
                filename=included_file._filename))

    def v2_playbook_on_play_start(self, play):
        self._play = play
        # Log an extra blank line to get space before each play
        self._log("")

        # the name of a play defaults to the hosts string
        name = play.get_name().strip()
        msg = u"PLAY [{name}]".format(name=name)

        self._log(msg)

    def v2_playbook_on_task_start(self, task, is_conditional):
        # Log an extra blank line to get space before each task
        self._log("")

        self._task = task

        if self._play.strategy != 'free':
            task_name = self._print_task_banner(task)
        else:
            task_name = task.get_name().strip()

        if task.loop is not None:
            # Don't try to stream from loops
            return
        if task.async_val:
            # Don't try to stream from async tasks
            return
        if task.action in STREAMING_ACTIONS:
            play_vars = self._play._variable_manager._hostvars

            hosts = self._get_task_hosts(task)
            for host, inventory_hostname in hosts:
                default_port = ACTION_LOG_STREAM_PORT.get(task.action)
                # This is intended to be only used for testing where
                # we change the port so we can run another instance
                # that doesn't conflict with one setup by the test
                # environment
                port = int(os.environ.get("ZUUL_CONSOLE_PORT", default_port))
                if port is None:
                    continue

                if (host in ('localhost', '127.0.0.1')):
                    # Don't try to stream from localhost
                    continue
                ip = play_vars[host].get(
                    'ansible_host', play_vars[host].get(
                        'ansible_inventory_host'))
                if (ip in ('localhost', '127.0.0.1')):
                    # Don't try to stream from localhost
                    continue
                # Note: this checks hostvars, but not task vars.
                if boolean(play_vars[host].get(
                        'zuul_console_disabled', False)):
                    # The user has told us not to even try
                    continue
                if play_vars[host].get('ansible_connection') in ('winrm',):
                    # The winrm connections don't support streaming for now
                    continue
                if play_vars[host].get('ansible_connection') in ('kubectl', ):
                    # Stream from the forwarded port on kubectl conns
                    if task.action in WINDOWS_ACTIONS:
                        port_id = 'stream_port2'
                    else:
                        port_id = 'stream_port1'
                    port = play_vars[host]['zuul']['resources'][
                        inventory_hostname].get(port_id)
                    if port is None:
                        self._log("[Zuul] Kubectl and socat must be installed "
                                  "on the Zuul executor for streaming output "
                                  "from pods")
                        continue
                    ip = '127.0.0.1'

                # Get a unique key for ZUUL_LOG_ID_MAP.  Use it to add
                # a counter to the log id so that if we run the same
                # task more than once, we get a unique log file.  See
                # comments in paths.py for details.
                log_host = paths._sanitize_filename(inventory_hostname)
                key = "%s-%s" % (self._task._uuid, log_host)
                count = paths.ZUUL_LOG_ID_MAP.get(key, 0) + 1
                paths.ZUUL_LOG_ID_MAP[key] = count
                log_id = "%s-%s-%s" % (
                    self._task._uuid, count, log_host)

                self._log("[%s] Starting to log %s for task %s"
                          % (host, log_id, task_name),
                          job=False, executor=True)
                streamer = Streamer(self, host, ip, port, log_id)
                streamer.start()
                self._streamers[self._task._uuid] = streamer

    def v2_playbook_on_handler_task_start(self, task):
        self.v2_playbook_on_task_start(task, False)

    def _stop_streamers(self):
        while True:
            keys = list(self._streamers.keys())
            if not keys:
                # No streamers, so all are stopped
                break
            streamer = self._streamers.pop(keys[0], None)
            if streamer is None:
                # Perhaps we raced; restart the loop and try again
                continue
            # Give 10 seconds of grace for the log to appear
            streamer.stop(10)
            # And 30 seconds to finish streaming after that
            streamer.join(30)
            if streamer.is_alive():
                msg = "[Zuul] Log Stream did not terminate"
                self._log(msg)

    def _stop_skipped_task_streamer(self, task):
        # Immediately stop a streamer for a skipped task.  We don't
        # expect a logfile in that situation.
        streamer = self._streamers.pop(task._uuid, None)
        if streamer is None:
            return
        # Give no grace since we don't expect the log to exist
        streamer.stop(0)
        # If the streamer connected, then we need to wait to join the
        # thread.  If not, then we can skip it because even if the
        # streamer is still attempting to connect, it will exit before
        # doing anything else.
        if streamer.connected:
            streamer.join(30)
            if streamer.is_alive():
                msg = "[Zuul] Log Stream did not terminate"
                self._log(msg)

    def _process_result_for_localhost(self, result, is_task=True):
        result_dict = dict(result._result)
        localhost_names = ('localhost', '127.0.0.1', '::1')
        is_localhost = False
        task_host = result._host.get_name()
        delegated_vars = result_dict.get('_ansible_delegated_vars', None)
        if delegated_vars:
            delegated_host = delegated_vars['ansible_host']
            if delegated_host in localhost_names:
                is_localhost = True
        elif result._task._variable_manager is None:
            # Handle fact gathering which doens't have a variable manager
            if task_host == 'localhost':
                is_localhost = True
        else:
            task_hostvars = result._task._variable_manager._hostvars[task_host]
            # Normally hosts in the inventory will have ansible_host
            # or ansible_inventory host defined.  The implied
            # inventory record for 'localhost' will have neither, so
            # default to that if none are supplied.
            if task_hostvars.get('ansible_host', task_hostvars.get(
                    'ansible_inventory_host', 'localhost')) in localhost_names:
                is_localhost = True

        if not is_localhost and is_task:
            self._stop_streamers()
        if result._task.action in ALL_ACTIONS:
            stdout_lines = zuul_filter_result(result_dict)
            # We don't have streaming for localhost, raw, or disabled
            # consoles, so get standard out after the fact.
            play_vars = self._play._variable_manager._hostvars
            # Note: this checks hostvars, but not task vars.
            is_disabled = boolean(play_vars[result._host.get_name()].get(
                'zuul_console_disabled', False))
            if (is_localhost or
                is_disabled or
                result._task.action in OUTPUT_ACTIONS):
                for line in stdout_lines:
                    hostname = self._get_hostname(result)
                    self._log("%s | %s " % (hostname, line))
                    self._check_failure_output(hostname, line)

    def _v2_runner_on_failed(self, result, ignore_errors=False):
        result_dict = dict(result._result)

        self._handle_exception(result_dict)

        if result_dict.get('msg') == 'All items completed':
            result_dict['status'] = 'ERROR'
            self._deferred_result = result_dict
            return

        self._process_result_for_localhost(result)

        def is_module_failure(msg):
            return isinstance(msg, str) and msg.startswith('MODULE FAILURE')

        if result._task.loop and 'results' in result_dict:
            # items have their own events
            pass
        elif is_module_failure(result_dict.get('msg', '')):
            self._log_module_failure(result, result_dict)
        else:
            self._log_message(
                result=result, status='ERROR', result_dict=result_dict)
        if ignore_errors:
            self._log_message(result, "Ignoring Errors", status="ERROR")

    def v2_runner_on_failed(self, result, ignore_errors=False):
        ret = self._v2_runner_on_failed(result, ignore_errors)
        if (not ignore_errors and
            not self.sent_failure_result and
            not is_rescuable(result._task)):
            self._result_logger.info("failure")
            self.sent_failure_result = True
        return ret

    def v2_runner_on_unreachable(self, result):
        ret = self._v2_runner_on_failed(result)
        self._result_logger.info("unreachable")
        return ret

    def v2_runner_on_skipped(self, result):
        self._stop_skipped_task_streamer(result._task)
        if result._task.loop:
            self._items_done = False
            self._deferred_result = dict(result._result)
        else:
            reason = result._result.get('skip_reason')
            if reason:
                # No reason means it's an item, which we'll log differently
                self._log_message(result, status='skipping', msg=reason)

    def v2_runner_item_on_skipped(self, result):
        reason = result._result.get('skip_reason')
        if reason:
            self._log_message(result, status='skipping', msg=reason)
        else:
            self._log_message(result, status='skipping')

        if self._deferred_result:
            self._process_deferred(result)

    def v2_runner_on_ok(self, result):
        result_dict = dict(result._result)

        if result._task.action in ('command', 'shell'):
            # The command module has a small set of msgs it returns;
            # we can use that to detect if decided not to execute the
            # command:
            # "Did/Would not run command since 'path' exists/does not exist"
            # is the message we're looking for.
            if 'not run command since' in result_dict.get('msg', ''):
                self._stop_skipped_task_streamer(result._task)

        if result._task.action in ('win_command', 'win_shell'):
            # The win_command module has a small set of msgs it returns;
            # we can use that to detect if decided not to execute the
            # command:
            # "skipped, since $creates exists" and "skipped, since
            # $removes does not exist" are the messages we're looking
            # for.
            m = result_dict.get('msg', '')
            if 'skipped, since' in m and 'exist' in m:
                self._stop_skipped_task_streamer(result._task)

        if (self._play.strategy == 'free'
                and self._last_task_banner != result._task._uuid):
            self._print_task_banner(result._task)

        self._clean_results(result_dict, result._task.action)
        if '_zuul_nolog_return' in result_dict:
            # We have a custom zuul module that doesn't want the parameters
            # from its returned splatted to stdout. This is typically for
            # modules that are collecting data to be displayed some other way.
            for key in list(result_dict.keys()):
                if key != 'changed':
                    result_dict.pop(key)

        if result_dict.get('changed', False):
            status = 'changed'
        else:
            status = 'ok'

        if (result_dict.get('msg') == 'All items completed'
                and not self._items_done):
            result_dict['status'] = status
            self._deferred_result = result_dict
            return

        if not result._task.loop:
            self._process_result_for_localhost(result)
        else:
            self._items_done = False

        self._handle_warnings(result_dict)

        if result._task.loop and 'results' in result_dict:
            # items have their own events
            pass
        elif to_text(result_dict.get('msg', '')).startswith('MODULE FAILURE'):
            self._log_module_failure(result, result_dict)
        elif result._task.action == 'debug':
            # this is a debug statement, handle it special
            for key in [k for k in result_dict
                        if k.startswith('_ansible')]:
                del result_dict[key]
            if 'changed' in result_dict:
                del result_dict['changed']
            keyname = next(iter(result_dict.keys()))
            # If it has msg, that means it was like:
            #
            #  debug:
            #    msg: Some debug text the user was looking for
            #
            # So we log it with self._log to get just the raw string the
            # user provided. Note that msg may be a multi line block quote
            # so we handle that here as well.
            if keyname == 'msg':
                msg_lines = to_text(result_dict['msg']).rstrip().split('\n')
                for msg_line in msg_lines:
                    self._log(msg=msg_line)
            else:
                self._log_message(
                    msg=json.dumps(result_dict, indent=2, sort_keys=True),
                    status=status, result=result)
        elif result._task.action not in STREAMING_ACTIONS:
            if 'msg' in result_dict:
                self._log_message(msg=result_dict['msg'],
                                  result=result, status=status)
            else:
                self._log_message(
                    result=result,
                    status=status)
        elif 'results' in result_dict:
            for res in result_dict['results']:
                self._log_message(
                    result,
                    "Runtime: {delta}".format(**res))
        elif result_dict.get('msg') == 'All items completed':
            self._log_message(result, result_dict['msg'])
        else:
            if 'delta' in result_dict:
                self._log_message(
                    result,
                    "Runtime: {delta}".format(
                        **result_dict))
            else:
                # NOTE(ianw) 2022-08-24 : *Fairly* sure that you only
                # fall into here when the call actually fails (and has
                # not start/end time), but it is ignored by
                # failed_when matching.
                self._log_message(result, msg='ERROR (ignored)',
                                  result_dict=result_dict)

    def v2_runner_item_on_ok(self, result):
        result_dict = dict(result._result)
        self._process_result_for_localhost(result, is_task=False)

        if result_dict.get('changed', False):
            status = 'changed'
        else:
            status = 'ok'

        # This fallback may not be strictly necessary. 'item' is the
        # default and we'll avoid problems in the common case if ansible
        # changes.
        loop_var = result_dict.get('ansible_loop_var', 'item')

        if to_text(result_dict.get('msg', '')).startswith('MODULE FAILURE'):
            self._log_module_failure(result, result_dict)
        elif result._task.action not in ALL_ACTIONS:
            if 'msg' in result_dict:
                self._log_message(
                    result=result, msg=result_dict['msg'], status=status)
            else:
                self._log_message(
                    result=result,
                    msg=json.dumps(result_dict[loop_var],
                                   indent=2, sort_keys=True),
                    status=status)
        else:
            stdout_lines = zuul_filter_result(result_dict)
            for line in stdout_lines:
                hostname = self._get_hostname(result)
                self._log("%s | %s " % (hostname, line))

            if isinstance(result_dict[loop_var], str):
                self._log_message(
                    result,
                    "Item: {loop_var} Runtime: {delta}".format(
                        loop_var=result_dict[loop_var],
                        delta=result_dict['delta']))
            else:
                self._log_message(
                    result,
                    "Item: Runtime: {delta}".format(
                        **result_dict))

        if self._deferred_result:
            self._process_deferred(result)

    def v2_runner_item_on_failed(self, result):
        result_dict = dict(result._result)
        self._process_result_for_localhost(result, is_task=False)

        # This fallback may not be strictly necessary. 'item' is the
        # default and we'll avoid problems in the common case if ansible
        # changes.
        loop_var = result_dict.get('ansible_loop_var', 'item')

        if to_text(result_dict.get('msg', '')).startswith('MODULE FAILURE'):
            self._log_module_failure(result, result_dict)
        elif result._task.action not in ALL_ACTIONS:
            self._log_message(
                result=result,
                msg="Item: {loop_var}".format(loop_var=result_dict[loop_var]),
                status='ERROR',
                result_dict=result_dict)
        else:
            stdout_lines = zuul_filter_result(result_dict)
            for line in stdout_lines:
                hostname = self._get_hostname(result)
                self._log("%s | %s " % (hostname, line))

            # self._log("Result: %s" % (result_dict))
            self._log_message(
                result, "Item: {loop_var} Result: {rc}".format(
                    loop_var=result_dict[loop_var], rc=result_dict['rc']))

        if self._deferred_result:
            self._process_deferred(result)

    def v2_playbook_on_stats(self, stats):
        # Add a spacer line before the stats so that there will be a line
        # between the last task and the recap
        self._log("")

        self._log("PLAY RECAP")

        hosts = sorted(stats.processed.keys())
        for host in hosts:
            t = stats.summarize(host)
            msg = (
                "{host} |"
                " ok: {ok}"
                " changed: {changed}"
                " unreachable: {unreachable}"
                " failed: {failures}".format(host=host, **t))

            # NOTE(pabelanger) Ansible 2.8 added rescued support
            if 'rescued' in t:
                # Even though skipped was in stable-2.7 and lower, only
                # stable-2.8 started rendering it. So just lump into rescued
                # check.
                msg += " skipped: {skipped} rescued: {rescued}".format(**t)
            # NOTE(pabelanger) Ansible 2.8 added ignored support
            if 'ignored' in t:
                msg += " ignored: {ignored}".format(**t)
            self._log(msg)

        # Add a spacer line after the stats so that there will be a line
        # between each playbook
        self._log("")

    def _process_deferred(self, result):
        self._items_done = True
        result_dict = self._deferred_result
        self._deferred_result = None
        status = result_dict.get('status')

        if status:
            self._log_message(result, "All items complete", status=status)

        # Log an extra blank line to get space after each task
        self._log("")

    def _print_task_banner(self, task):

        task_name = task.get_name().strip()

        if task.loop:
            task_type = 'LOOP'
        else:
            task_type = 'TASK'

        # TODO(mordred) With the removal of printing task args, do we really
        # want to keep doing this section?
        task_args = task.args.copy()
        is_shell = task_args.pop('_uses_shell', False)
        if is_shell and task_name == 'command':
            task_name = 'shell'
        # win_shell doesn't use _uses_shell.
        raw_params = task_args.pop('_raw_params', '').split('\n')
        # If there's just a single line, go ahead and print it
        if len(raw_params) == 1 and task_name in STREAMING_ACTIONS:
            task_name = '{name}: {command}'.format(
                name=task_name, command=raw_params[0])

        msg = "{task_type} [{task}]".format(
            task_type=task_type,
            task=task_name)
        self._log(msg)
        return task

    def _get_task_hosts(self, task):
        result = []

        # _restriction returns the parsed/compiled list of hosts after
        # applying subsets/limits
        hosts = self._play._variable_manager._inventory._restriction
        for inventory_host in hosts:
            # If this task has as delegate to, we don't care about the play
            # hosts, we care about the task's delegate target.
            if task.delegate_to:
                host = task.delegate_to
            else:
                host = inventory_host
            result.append((host, inventory_host))

        return result

    def _dump_result_dict(self, result_dict):
        result_dict = result_dict.copy()
        for key in list(result_dict.keys()):
            if key.startswith('_ansible'):
                del result_dict[key]
        zuul_filter_result(result_dict)
        return result_dict

    def _log_message(self, result, msg=None, status="ok", result_dict=None):
        hostname = self._get_hostname(result)
        if result_dict:
            result_dict = self._dump_result_dict(result_dict)
        if result._task.no_log:
            self._log("{host} | {msg}".format(
                host=hostname,
                msg="Output suppressed because no_log was given"))
            return
        if (not msg and result_dict
                and set(result_dict.keys()) == set(['msg', 'failed'])):
            msg = result_dict['msg']
            result_dict = None
        if msg:
            # ensure msg is a string; e.g.
            #
            # debug:
            #   msg: '{{ var }}'
            #
            # may not be!
            msg_lines = to_text(msg).rstrip().split('\n')
            if len(msg_lines) > 1:
                self._log("{host} | {status}:".format(
                    host=hostname, status=status))
                for msg_line in msg_lines:
                    self._log("{host} | {msg_line}".format(
                        host=hostname, msg_line=msg_line))
            else:
                self._log("{host} | {status}: {msg}".format(
                    host=hostname, status=status, msg=msg))
        else:
            self._log("{host} | {status}".format(
                host=hostname, status=status))
        if result_dict:
            result_string = json.dumps(result_dict, indent=2, sort_keys=True)
            for line in result_string.split('\n'):
                self._log("{host} | {line}".format(host=hostname, line=line))

    def _get_hostname(self, result):
        delegated_vars = result._result.get('_ansible_delegated_vars', None)
        if delegated_vars:
            return "{host} -> {delegated_host}".format(
                host=result._host.get_name(),
                delegated_host=delegated_vars['ansible_host'])
        else:
            return result._host.get_name()



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from tests.base import ZuulTestCase, ZuulGithubAppTestCase
from zuul.driver.zuul.zuulmodel import ZuulTriggerEvent


class TestZuulTriggerParentChangeEnqueued(ZuulTestCase):
    tenant_config_file = 'config/zuultrigger/parent-change-enqueued/main.yaml'

    def test_zuul_trigger_parent_change_enqueued(self):
        "Test Zuul trigger event: parent-change-enqueued"
        # This test has the following three changes:
        # B1 -> A; B2 -> A
        # When A is enqueued in the gate, B1 and B2 should both attempt
        # to be enqueued in both pipelines.  B1 should end up in check
        # and B2 in gate because of differing pipeline requirements.
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B1 = self.fake_gerrit.addFakeChange('org/project', 'master', 'B1')
        B2 = self.fake_gerrit.addFakeChange('org/project', 'master', 'B2')
        A.addApproval('Code-Review', 2)
        B1.addApproval('Code-Review', 2)
        B2.addApproval('Code-Review', 2)
        A.addApproval('Verified', 1, username="for-check")   # reqd by check
        A.addApproval('Verified', 1, username="for-gate")    # reqd by gate
        B1.addApproval('Verified', 1, username="for-check")  # go to check
        B2.addApproval('Verified', 1, username="for-gate")   # go to gate
        B1.addApproval('Approved', 1)
        B2.addApproval('Approved', 1)
        B1.setDependsOn(A, 1)
        B2.setDependsOn(A, 1)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        # Jobs are being held in build to make sure that 3,1 has time
        # to enqueue behind 1,1 so that the test is more
        # deterministic.
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(len(self.history), 3)
        for job in self.history:
            if job.changes == '1,1':
                self.assertEqual(job.name, 'project-gate')
            elif job.changes == '1,1 2,1':
                self.assertEqual(job.name, 'project-check')
            elif job.changes == '1,1 3,1':
                self.assertEqual(job.name, 'project-gate')
            else:
                raise Exception("Unknown job")

        # Now directly enqueue a change into the check. As no pipeline reacts
        # on parent-change-enqueued from pipeline check no
        # parent-change-enqueued event is expected.
        _add_trigger_event = self.scheds.first.sched.addTriggerEvent

        def addTriggerEvent(driver_name, event):
            self.assertNotIsInstance(event, ZuulTriggerEvent)
            _add_trigger_event(driver_name, event)

        with mock.patch.object(
            self.scheds.first.sched, "addTriggerEvent", addTriggerEvent
        ):
            C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
            C.addApproval('Verified', 1, username="for-check")
            D = self.fake_gerrit.addFakeChange('org/project', 'master', 'D')
            D.addApproval('Verified', 1, username="for-check")
            D.setDependsOn(C, 1)
            self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(1))

            self.waitUntilSettled()
            self.assertEqual(len(self.history), 4)


class TestZuulTriggerParentChangeEnqueuedGithub(ZuulGithubAppTestCase):
    tenant_config_file = \
        'config/zuultrigger/parent-change-enqueued-github/main.yaml'
    config_file = 'zuul-github-driver.conf'

    def test_zuul_trigger_parent_change_enqueued(self):
        "Test Zuul trigger event: parent-change-enqueued"
        # This test has the following three changes:
        # B1 -> A; B2 -> A
        # When A is enqueued in the gate, B1 and B2 should both attempt
        # to be enqueued in both pipelines.  B1 should end up in check
        # and B2 in gate because of differing pipeline requirements.
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_github.openFakePullRequest('org/project', 'master', 'A')
        msg = "Depends-On: https://github.com/org/project/pull/%s" % A.number
        B1 = self.fake_github.openFakePullRequest(
            'org/project', 'master', 'B1', body=msg)
        B2 = self.fake_github.openFakePullRequest(
            'org/project', 'master', 'B2', body=msg)
        A.addReview('derp', 'APPROVED')
        B1.addReview('derp', 'APPROVED')
        B2.addReview('derp', 'APPROVED')
        A.addLabel('for-gate')    # required by gate
        A.addLabel('for-check')   # required by check
        B1.addLabel('for-check')  # should go to check
        B2.addLabel('for-gate')   # should go to gate

        # In this case we have two installations
        # 1: org/common-config, org/project (used by tenant-one and tenant-two)
        # 2: org2/project (only used by tenant-two)
        # In order to track accesses to the installations enable client
        # recording in the fake github.
        self.fake_github.record_clients = True

        self.fake_github.emitEvent(A.getReviewAddedEvent('approved'))
        # Jobs are being held in build to make sure that 3,1 has time
        # to enqueue behind 1,1 so that the test is more
        # deterministic.
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(len(self.history), 3)
        for job in self.history:
            if job.changes == '1,{}'.format(A.head_sha):
                self.assertEqual(job.name, 'project-gate')
            elif job.changes == '1,{} 2,{}'.format(A.head_sha, B1.head_sha):
                self.assertEqual(job.name, 'project-check')
            elif job.changes == '1,{} 3,{}'.format(A.head_sha, B2.head_sha):
                self.assertEqual(job.name, 'project-gate')
            else:
                raise Exception("Unknown job")

        # Now directly enqueue a change into the check. As no pipeline reacts
        # on parent-change-enqueued from pipeline check no
        # parent-change-enqueued event is expected.
        self.waitUntilSettled()

        _add_trigger_event = self.scheds.first.sched.addTriggerEvent

        def addTriggerEvent(driver_name, event):
            self.assertNotIsInstance(event, ZuulTriggerEvent)
            _add_trigger_event(driver_name, event)

        with mock.patch.object(
            self.scheds.first.sched, "addTriggerEvent", addTriggerEvent
        ):
            C = self.fake_github.openFakePullRequest(
                'org/project', 'master', 'C'
            )
            C.addLabel('for-check')  # should go to check

            msg = "Depends-On: https://github.com/org/project1/pull/{}".format(
                C.number
            )
            D = self.fake_github.openFakePullRequest(
                'org/project', 'master', 'D', body=msg)
            D.addLabel('for-check')  # should go to check
            self.fake_github.emitEvent(C.getPullRequestOpenedEvent())

            self.waitUntilSettled()
            self.assertEqual(len(self.history), 4)

        # After starting recording installation containing org2/project
        # should not be contacted
        gh_manager = self.fake_github._github_client_manager
        inst_id_to_check = gh_manager.installation_map['org2/project']
        inst_clients = [x for x in gh_manager.recorded_clients
                        if x._inst_id == inst_id_to_check]
        self.assertEqual(len(inst_clients), 0)


class TestZuulTriggerProjectChangeMerged(ZuulTestCase):

    tenant_config_file = 'config/zuultrigger/project-change-merged/main.yaml'

    def test_zuul_trigger_project_change_merged(self):
        # This test has the following three changes:
        # A, B, C;  B conflicts with A, but C does not.
        # When A is merged, B and C should be checked for conflicts,
        # and B should receive a -1.
        # D and E are used to repeat the test in the second part, but
        # are defined here to that they end up in the trigger cache.
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        D = self.fake_gerrit.addFakeChange('org/project', 'master', 'D')
        E = self.fake_gerrit.addFakeChange('org/project', 'master', 'E')
        A.addPatchset({'conflict': 'foo'})
        B.addPatchset({'conflict': 'bar'})
        D.addPatchset({'conflict2': 'foo'})
        E.addPatchset({'conflict2': 'bar'})
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project-gate')
        self.assertEqual(A.reported, 2)
        self.assertEqual(B.reported, 1)
        self.assertEqual(C.reported, 0)
        self.assertEqual(D.reported, 0)
        self.assertEqual(E.reported, 0)
        self.assertIn(
            "Merge Failed.\n\nThis change or one of its cross-repo "
            "dependencies was unable to be automatically merged with the "
            "current state of its repository. Please rebase the change and "
            "upload a new patchset.",
            B.messages[0])
        self.assertIn(
            'Error merging gerrit/org/project for 2,2',
            B.messages[0])

        self.assertTrue("project:{org/project} status:open" in
                        self.fake_gerrit.queries)

        # Ensure the gerrit driver has updated its cache after the
        # previous comments were left:
        self.fake_gerrit.addEvent(A.getChangeCommentEvent(2))
        self.fake_gerrit.addEvent(B.getChangeCommentEvent(2))
        self.waitUntilSettled()

        # Reconfigure and run the test again.  This is a regression
        # check to make sure that we don't end up with a stale trigger
        # cache that has references to projects from the old
        # configuration.
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))

        D.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(D.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, 'project-gate')
        self.assertEqual(A.reported, 2)
        self.assertEqual(B.reported, 1)
        self.assertEqual(C.reported, 0)
        self.assertEqual(D.reported, 2)
        self.assertEqual(E.reported, 1)
        self.assertIn(
            "Merge Failed.\n\nThis change or one of its cross-repo "
            "dependencies was unable to be automatically merged with the "
            "current state of its repository. Please rebase the change and "
            "upload a new patchset.",
            E.messages[0])
        self.assertIn(
            'Error merging gerrit/org/project for 5,2',
            E.messages[0])
        self.assertIn("project:{org/project} status:open",
                      self.fake_gerrit.queries)



# Copyright 2019 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import concurrent.futures
import configparser
import json
import logging
import os
import shutil
import subprocess
import sys
import zuul.ansible
import importlib.resources

from zuul.lib.config import get_default


class ManagedAnsible:
    log = logging.getLogger('zuul.managed_ansible')

    def __init__(self, config, version, runtime_install_root=None):
        self.version = version

        requirements = get_default(config, version, 'requirements')
        self._requirements = requirements.split(' ')

        common_requirements = get_default(config, 'common', 'requirements')
        if common_requirements:
            self._requirements.extend(common_requirements.split(' '))

        self.deprecated = get_default(config, version, 'deprecated', False)

        self._ansible_roots = [os.path.join(
            sys.exec_prefix, 'lib', 'zuul', 'ansible')]
        if runtime_install_root:
            self._ansible_roots.append(runtime_install_root)

        self.install_root = self._ansible_roots[-1]

    def ensure_ansible(self, upgrade=False):
        self.log.info('Installing ansible %s, requirements: %s, '
                      'extra packages: %s',
                      self.version, self._requirements, self.extra_packages)
        self._run_pip(self._requirements + self.extra_packages,
                      upgrade=upgrade)

    def _run_pip(self, requirements, upgrade=False):
        cmd = [os.path.join(self.venv_path, 'bin', 'pip'), 'install',
               '--no-cache-dir']
        if upgrade:
            cmd.append('-U')
        cmd.extend(requirements)
        self.log.debug('Running pip: %s', ' '.join(cmd))

        p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if p.returncode != 0:
            raise Exception('Package installation failed with exit code %s '
                            'during processing ansible %s:\n'
                            'stdout:\n%s\n'
                            'stderr:\n%s' % (p.returncode, self.version,
                                             p.stdout.decode(),
                                             p.stderr.decode()))
        self.log.debug('Successfully installed packages %s', requirements)

    def ensure_venv(self):
        if self.python_path:
            self.log.debug(
                'Virtual environment %s already existing', self.venv_path)
            return

        venv_path = os.path.join(self.install_root, self.version)
        self.log.info('Creating venv %s', venv_path)

        python_executable = sys.executable
        if hasattr(sys, 'real_prefix'):
            # We're inside a virtual env and the venv module behaves strange
            # if we're calling it from there so default to
            # <real_prefix>/bin/python3
            python_executable = os.path.join(sys.real_prefix, 'bin', 'python3')

        # We don't use directly the venv module here because its behavior is
        # broken if we're already in a virtual environment.
        cmd = [sys.executable, '-m', 'virtualenv',
               '-p', python_executable, venv_path]
        p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if p.returncode != 0:
            raise Exception('venv creation failed with exit code %s:\n'
                            'stdout:\n%s\n'
                            'stderr:\n%s' % (p.returncode, p.stdout.decode(),
                                             p.stderr.decode()))

    @property
    def venv_path(self):
        for root in reversed(self._ansible_roots):
            # Check user configured paths first
            venv_path = os.path.join(root, self.version)
            if os.path.exists(venv_path):
                return venv_path
        return None

    @property
    def python_path(self):
        venv_path = self.venv_path
        if venv_path:
            return os.path.join(self.venv_path, 'bin', 'python')
        return None

    @property
    def extra_packages(self):
        mapping = str.maketrans({
            '.': None,
            '-': '_',
        })
        env_var = 'ANSIBLE_%s_EXTRA_PACKAGES' % self.version.upper().translate(
            mapping)

        packages = os.environ.get(env_var)
        result = []
        if packages:
            result.extend(packages.strip().split(' '))

        common_packages = os.environ.get('ANSIBLE_EXTRA_PACKAGES')
        if common_packages:
            result.extend(common_packages.strip().split(' '))

        return result

    def __repr__(self):
        return 'Ansible {a.version}, {a.deprecated}'.format(
            a=self)


class AnsibleManager:
    log = logging.getLogger('zuul.ansible_manager')

    def __init__(self, zuul_ansible_dir=None, default_version=None,
                 runtime_install_root=None):
        self._supported_versions = {}
        self.default_version = None
        self.zuul_ansible_dir = zuul_ansible_dir
        self.runtime_install_root = runtime_install_root

        self.load_ansible_config()

        # If configured, override the default version
        if default_version:
            self.requestVersion(default_version)
            self.default_version = default_version

    def load_ansible_config(self):
        ref = importlib.resources.files('zuul').joinpath(
            'lib/ansible-config.conf')
        c = ref.read_bytes().decode()
        config = configparser.ConfigParser()
        config.read_string(c)

        for version in config.sections():
            # The common section is no ansible version
            if version == 'common':
                continue

            ansible = ManagedAnsible(
                config, version,
                runtime_install_root=self.runtime_install_root)

            if ansible.version in self._supported_versions:
                raise RuntimeError(
                    'Ansible version %s already defined' % ansible.version)

            self._supported_versions[ansible.version] = ansible

        default_version = get_default(
            config, 'common', 'default_version', None)
        if not default_version:
            raise RuntimeError('A default ansible version must be specified')

        # Validate that this version is known
        self._getAnsible(default_version)
        self.default_version = default_version

    def install(self, upgrade=False):
        # virtualenv sets up a shared directory of pip seed packages per
        # python version. If we run virtualenv in parallel we can have one
        # create the dirs but not be finished filling them with content, a
        # second notice the dir is there so it just goes on with what it's
        # done, and thus races leaving us with virtualenvs minus pip.
        for a in self._supported_versions.values():
            a.ensure_venv()
        # Note: With higher number of threads pip seems to have some race
        # leading to occasional failures during setup of all ansible
        # environments. Thus we limit the number of workers to reduce the risk
        # of hitting this race.
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = {executor.submit(a.ensure_ansible, upgrade): a
                       for a in self._supported_versions.values()}
            for future in concurrent.futures.as_completed(futures):
                future.result()

    def _validate_ansible(self, version):
        result = True
        try:
            command = [
                self.getAnsibleCommand(version, 'ansible'),
                '--version',
            ]

            ret = subprocess.run(command,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 check=True)
            self.log.info('Ansible version %s information: \n%s',
                          version, ret.stdout.decode())
        except subprocess.CalledProcessError:
            result = False
            self.log.exception("Ansible version %s not working" % version)
        except Exception:
            result = False
            self.log.exception(
                'Ansible version %s not installed' % version)
        return result

    def _validate_packages(self, version):
        result = False
        try:
            extra_packages = self._getAnsible(version).extra_packages
            if not extra_packages:
                return True

            # Formerly this used pkg_resources which has been deprecated. If
            # there is a better way to accomplish this task please change
            # this approach.
            # Ask pip to resolve a dry run installation to determine if any new
            # packages need to be installed. A json doc is emitted to stdout
            # including a list of things to install if necessary.
            command = [self.getAnsibleCommand(version, 'pip'), 'install',
                       '--dry-run', '--quiet', '--report', '-']
            command += extra_packages
            ret = subprocess.run(command,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            if ret.stdout:
                to_be_installed = json.loads(ret.stdout)["install"]
            else:
                to_be_installed = None
            # We check manually so that we can log the missing packages
            # properly. We also need to check the JSON output to determine
            # if any changes were necessary.
            if to_be_installed:
                missing = ["%s %s" %
                           (x["metadata"]["name"], x["metadata"]["version"])
                           for x in to_be_installed]
                self.log.error(
                    'Ansible version %s installation is missing packages %s',
                    version, missing)
            else:
                result = True
        except Exception:
            self.log.exception(
                'Exception checking Ansible version %s packages' %
                version)
        return result

    def validate(self):
        result = True
        for version in self._supported_versions:
            if not self._validate_ansible(version):
                result = False
            elif not self._validate_packages(version):
                result = False
        return result

    def _getAnsible(self, version):
        if not version:
            version = self.default_version

        ansible = self._supported_versions.get(version)
        if not ansible:
            raise Exception('Requested ansible version %s not found' % version)
        return ansible

    def getAnsibleCommand(self, version, command='ansible-playbook'):
        ansible = self._getAnsible(version)
        venv_path = ansible.venv_path
        if not venv_path:
            raise Exception('Requested ansible version \'%s\' is not '
                            'installed' % version)
        return os.path.join(ansible.venv_path, 'bin', command)

    def getAnsibleInstallDir(self, version):
        ansible = self._getAnsible(version)
        venv_path = ansible.venv_path
        if not venv_path:
            raise Exception('Requested ansible version \'%s\' is not '
                            'installed' % version)
        return venv_path

    def getAnsibleDir(self, version):
        ansible = self._getAnsible(version)
        return os.path.join(self.zuul_ansible_dir, ansible.version)

    def getAnsiblePluginDir(self, version):
        return os.path.join(self.getAnsibleDir(version), 'zuul', 'ansible')

    def requestVersion(self, version):
        if version not in self._supported_versions:
            raise Exception(
                'Requested ansible version \'%s\' is unknown. Supported '
                'versions are %s' % (
                    version, ', '.join(self._supported_versions)))

    def getSupportedVersions(self):
        versions = []
        for version in self._supported_versions:
            versions.append((version, version == self.default_version))
        return versions

    def copyAnsibleFiles(self):
        if os.path.exists(self.zuul_ansible_dir):
            # Ensure we can delete the files by setting writtable mode
            for dirpath, dirnames, filenames in os.walk(self.zuul_ansible_dir):
                os.chmod(dirpath, 0o755)
                for filename in filenames:
                    os.chmod(os.path.join(dirpath, filename), 0o600)
            shutil.rmtree(self.zuul_ansible_dir)

        library_path = os.path.dirname(os.path.abspath(zuul.ansible.__file__))
        for ansible in self._supported_versions.values():
            ansible_dir = os.path.join(self.zuul_ansible_dir, ansible.version)
            plugin_dir = os.path.join(ansible_dir, 'zuul', 'ansible')
            source_path = os.path.join(library_path, ansible.version)

            os.makedirs(plugin_dir, exist_ok=True)
            for fn in os.listdir(source_path):
                if fn in ('__pycache__', 'base'):
                    continue
                full_path = os.path.join(source_path, fn)
                if os.path.isdir(full_path):
                    shutil.copytree(full_path, os.path.join(plugin_dir, fn))
                else:
                    shutil.copy(os.path.join(source_path, fn), plugin_dir)

            # We're copying zuul.ansible.* into a directory we are going
            # to add to pythonpath, so our plugins can "import
            # zuul.ansible".  But we're not installing all of zuul, so
            # create a __init__.py file for the stub "zuul" module.
            module_paths = [
                os.path.join(ansible_dir, 'zuul'),
                os.path.join(ansible_dir, 'zuul', 'ansible'),
            ]
            for fn in module_paths:
                with open(os.path.join(fn, '__init__.py'), 'w'):
                    # Nothing to do here, we just want the file to exist.
                    pass



# Copyright 2014 OpenStack Foundation
# Copyright 2021-2022, 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import socket
import sys
import threading
import time
from abc import ABCMeta
from configparser import ConfigParser
from contextlib import nullcontext

from kazoo.exceptions import NoNodeError
from kazoo.retry import KazooRetry

from zuul.lib import commandsocket
from zuul.lib.config import get_default
from zuul.lib.logutil import get_annotated_logger
from zuul.lib import tracing
from zuul.merger import merger
from zuul.model import (
    FilesChangesCompletedEvent, MergeCompletedEvent, MergeRequest
)
from zuul.lib.monitoring import MonitoringServer
from zuul.version import get_version_string
from zuul.zk import ZooKeeperClient
from zuul.zk.components import MergerComponent, COMPONENT_REGISTRY
from zuul.zk.event_queues import PipelineResultEventQueue
from zuul.zk.merger import MergerApi

COMMANDS = [
    commandsocket.StopCommand,
    commandsocket.PauseCommand,
    commandsocket.UnPauseCommand,
    commandsocket.GracefulCommand
]


class BaseRepoLocks(metaclass=ABCMeta):

    def getRepoLock(self, connection_name, project_name):
        return nullcontext()


class RepoLocks(BaseRepoLocks):

    def __init__(self):
        self.locks = {}

    def getRepoLock(self, connection_name, project_name):
        key = '%s:%s' % (connection_name, project_name)
        self.locks.setdefault(key, threading.Lock())
        return self.locks[key]


class BaseMergeServer(metaclass=ABCMeta):
    log = logging.getLogger("zuul.BaseMergeServer")

    _repo_locks_class = BaseRepoLocks
    _merger_api_class = MergerApi

    def __init__(
        self,
        config: ConfigParser,
        component: str,
        connections,
    ):
        self.connections = connections
        self._merger_running = False
        self._merger_paused = False
        self.merge_email = get_default(config, 'merger', 'git_user_email',
                                       'zuul.merger.default@example.com')
        self.merge_name = get_default(config, 'merger', 'git_user_name',
                                      'Zuul Merger Default')
        self.merge_speed_limit = get_default(
            config, 'merger', 'git_http_low_speed_limit', '1000')
        self.merge_speed_time = get_default(
            config, 'merger', 'git_http_low_speed_time', '30')
        self.git_timeout = get_default(config, 'merger', 'git_timeout', 300)

        self.merge_root = get_default(config, component, 'git_dir',
                                      '/var/lib/zuul/{}-git'.format(component))

        self.config = config

        self.tracing = tracing.Tracing(self.config)
        self.zk_client = ZooKeeperClient.fromConfig(self.config)
        self.zk_client.connect()

        self.result_events = PipelineResultEventQueue.createRegistry(
            self.zk_client
        )

        self.merger_thread = threading.Thread(
            target=self.runMerger,
            name="Merger",
        )

        self.merger_loop_wake_event = threading.Event()

        self.merger_api = self._merger_api_class(
            self.zk_client,
            merge_request_callback=self.merger_loop_wake_event.set,
        )

        # This merger and its git repos are used to maintain
        # up-to-date copies of all the repos that are used by jobs, as
        # well as to support the merger:cat functon to supply
        # configuration information to Zuul when it starts.
        self.merger = self._getMerger(self.merge_root, None,
                                      execution_context=False)

        # Repo locking is needed on the executor
        self.repo_locks = self._repo_locks_class()

    def _getMerger(self, root, cache_root, logger=None,
                   execution_context=True, scheme=None,
                   cache_scheme=None):
        return merger.Merger(
            root,
            self.connections,
            self.zk_client,
            self.merge_email,
            self.merge_name,
            self.merge_speed_limit,
            self.merge_speed_time,
            cache_root,
            logger,
            execution_context=execution_context,
            git_timeout=self.git_timeout,
            scheme=scheme,
            cache_scheme=cache_scheme,
        )

    def _repoLock(self, connection_name, project_name):
        # The merger does not need locking so return a null lock.
        return nullcontext()

    def _update(self, connection_name, project_name, zuul_event_id=None):
        # The executor overrides _update so it can do the update
        # asynchronously.
        self.merger.updateRepo(connection_name, project_name,
                               zuul_event_id=zuul_event_id)

    def start(self):
        self.log.debug('Starting merger')
        self._merger_running = True
        self.merger_thread.start()

    def stop(self):
        self.log.debug('Stopping base merge server')
        self._merger_running = False
        self.merger_loop_wake_event.set()
        self.tracing.stop()
        self.log.debug('Stopped base merge server')

    def join(self):
        self.log.debug('Joining base merge server')
        self.merger_loop_wake_event.set()
        self.merger_thread.join()
        self.zk_client.disconnect()
        self.log.debug('Joined base merge server')

    def pause(self):
        self.log.debug('Pausing merger')
        self._merger_paused = True

    def unpause(self):
        self.log.debug('Resuming merger')
        self._merger_paused = False
        self.merger_loop_wake_event.set()

    def isComponentRunning(self):
        return self._merger_running

    def runMerger(self):
        while self._merger_running:
            self.merger_loop_wake_event.wait()
            self.merger_loop_wake_event.clear()
            if self._merger_paused:
                continue
            try:
                for merge_request in self.merger_api.next():
                    if not self._merger_running:
                        break

                    self._lockAndRunMergeJob(merge_request)
            except Exception:
                self.log.exception("Error in merge thread:")
                time.sleep(5)
                self.merger_loop_wake_event.set()

    def _lockAndRunMergeJob(self, merge_request):
        # Lock and update the merge request
        if not self.merger_api.lock(merge_request, blocking=False):
            return
        with tracing.startSpanInContext(
                merge_request.span_context, "MergerJob",
                attributes={"merger": self.hostname}):
            self._runMergeJob(merge_request)

    def _runMergeJob(self, merge_request):
        log = get_annotated_logger(
            self.log, merge_request.event_id
        )

        # Ensure that the request is still in state requested. This method is
        # called based on cached data and there might be a mismatch between the
        # cached state and the real state of the request. The lock might
        # have been successful because the request is already completed and
        # thus unlocked.
        if merge_request.state != MergeRequest.REQUESTED:
            self._retry(merge_request.lock, self.merger_api.unlock,
                        merge_request)

        try:
            merge_request.state = MergeRequest.RUNNING
            params = self.merger_api.getParams(merge_request)
        except Exception:
            log.exception("Exception while preparing to start merge job")
            # If we failed at this point, we have not written anything
            # to ZK yet; the only thing we need to do is to ensure
            # that we release the lock, and another merger will be
            # able to grab the request.
            self._retry(merge_request.lock, self.merger_api.unlock,
                        merge_request)
            return

        result = None
        try:
            self.merger_api.clearParams(merge_request)
            # Directly update the merge request in ZooKeeper, so we
            # don't loop over and try to lock it again and again.
            self.merger_api.update(merge_request)
            self.log.debug("Next executed merge job: %s", merge_request)
            try:
                result = self.executeMergeJob(merge_request, params)
            except Exception:
                log.exception("Error running merge job:")
        finally:
            self.completeMergeJob(merge_request, result)

    def executeMergeJob(self, merge_request, params):
        result = None
        start = time.monotonic()
        if merge_request.job_type == MergeRequest.MERGE:
            result = self.merge(merge_request, params)
        elif merge_request.job_type == MergeRequest.CAT:
            result = self.cat(merge_request, params)
        elif merge_request.job_type == MergeRequest.REF_STATE:
            result = self.refstate(merge_request, params)
        elif merge_request.job_type == MergeRequest.FILES_CHANGES:
            result = self.fileschanges(merge_request, params)
        end = time.monotonic()
        result['elapsed_time'] = end - start
        return result

    def cat(self, merge_request, args):
        self.log.debug("Got cat job: %s", merge_request.uuid)

        connection_name = args['connection']
        project_name = args['project']

        lock = self.repo_locks.getRepoLock(connection_name, project_name)
        try:
            self._update(connection_name, project_name)
            with lock:
                files = self.merger.getFiles(connection_name, project_name,
                                             args['branch'], args['files'],
                                             args.get('dirs'))
        except Exception:
            result = dict(update=False)
        else:
            result = dict(updated=True, files=files)

        return result

    def merge(self, merge_request, args):
        self.log.debug("Got merge job: %s", merge_request.uuid)
        zuul_event_id = merge_request.event_id

        errors = []
        result = dict(merged=False,
                      commit=None,
                      files=None,
                      repo_state=None,
                      zuul_event_id=zuul_event_id,
                      errors=errors)
        for item in args['items']:
            try:
                self._update(item['connection'], item['project'])
            except Exception:
                msg = (f"Unable to update {item['connection']}/"
                       f"{item['project']}")
                errors.append(msg)
                return result

        ret = self.merger.mergeChanges(
            args['items'], args.get('files'),
            args.get('dirs', []),
            args.get('repo_state'),
            branches=args.get('branches'),
            repo_locks=self.repo_locks,
            zuul_event_id=zuul_event_id,
            errors=errors)

        if ret is not None:
            result['merged'] = True
            (result['commit'], result['files'], result['repo_state'],
             recent, orig_commit, ops) = ret
        return result

    def refstate(self, merge_request, args):
        self.log.debug("Got refstate job: %s", merge_request.uuid)
        zuul_event_id = merge_request.event_id
        success, repo_state, item_in_branches = \
            self.merger.getRepoState(
                args['items'], self.repo_locks, branches=args.get('branches'))
        result = dict(updated=success,
                      repo_state=repo_state,
                      item_in_branches=item_in_branches)
        result['zuul_event_id'] = zuul_event_id
        return result

    def fileschanges(self, merge_request, args):
        self.log.debug("Got fileschanges job: %s", merge_request.uuid)
        zuul_event_id = merge_request.event_id
        changes = args.get('changes')
        results = []
        for change in changes:
            connection_name = change['connection']
            project_name = change['project']

            lock = self.repo_locks.getRepoLock(connection_name, project_name)
            try:
                self._update(connection_name, project_name,
                             zuul_event_id=zuul_event_id)
                with lock:
                    files = self.merger.getFilesChanges(
                        connection_name, project_name,
                        change['branch'], change['tosha'],
                        zuul_event_id=zuul_event_id)
                results.append(files)
            except Exception:
                return dict(updated=False)

        return dict(updated=True, files=results)

    def completeMergeJob(self, merge_request, result):
        log = get_annotated_logger(self.log, merge_request.event_id)

        # Always provide a result event, even if we have no
        # information; otherwise items can get stuck in the queue.
        if result is None:
            result = {}

        payload = json.dumps(result, sort_keys=True)
        self.log.debug("Completed %s job %s: payload size: %s",
                       merge_request.job_type, merge_request.uuid,
                       sys.getsizeof(payload))
        merged = result.get("merged", False)
        updated = result.get("updated", False)
        commit = result.get("commit")
        repo_state = result.get("repo_state", {})
        item_in_branches = result.get("item_in_branches", [])
        files = result.get("files", {})
        errors = result.get("errors", [])
        elapsed_time = result.get("elapsed_time")

        log.info(
            "Merge %s complete, merged: %s, updated: %s, commit: %s, "
            "branches: %s",
            merge_request,
            merged,
            updated,
            commit,
            item_in_branches,
        )

        lock_valid = merge_request.lock.is_still_valid()
        if not lock_valid:
            return

        # Provide a result either via a result future or a result event
        if merge_request.result_path:
            log.debug(
                "Providing synchronous result via future for %s",
                merge_request,
            )
            self._retry(merge_request.lock,
                        self.merger_api.reportResult, merge_request, result)

        elif merge_request.build_set_uuid:
            log.debug(
                "Providing asynchronous result via result event for %s",
                merge_request,
            )
            if merge_request.job_type == MergeRequest.FILES_CHANGES:
                event = FilesChangesCompletedEvent(
                    merge_request.uuid,
                    merge_request.build_set_uuid,
                    files,
                    elapsed_time,
                    merge_request.span_info,
                )
            else:
                event = MergeCompletedEvent(
                    merge_request.uuid,
                    merge_request.build_set_uuid,
                    merged,
                    updated,
                    commit,
                    files,
                    repo_state,
                    item_in_branches,
                    errors,
                    elapsed_time,
                    merge_request.span_info,
                    merge_request.event_id,
                )

            def put_complete_event(log, merge_request, event):
                try:
                    self.result_events[merge_request.tenant_name][
                        merge_request.pipeline_name].put(event)
                except NoNodeError:
                    log.warning("Pipeline was removed: %s",
                                merge_request.pipeline_name)

            self._retry(merge_request.lock,
                        put_complete_event, log, merge_request, event)

        # Set the merge request to completed, unlock and delete it. Although
        # the state update is mainly for consistency reasons, it might come in
        # handy in case the deletion or unlocking failes. Thus, we know that
        # the merge request was already processed and we have a result in the
        # result queue.
        merge_request.state = MergeRequest.COMPLETED
        self._retry(merge_request.lock,
                    self.merger_api.update, merge_request)
        self._retry(merge_request.lock,
                    self.merger_api.unlock, merge_request)
        # TODO (felix): If we want to optimize ZK requests, we could only call
        # the remove() here.
        self.merger_api.remove(merge_request)

    def _retry(self, lock, func, *args, **kw):
        """Retry a method to deal with ZK connection issues

        This is a helper method to retry retryable ZK operations as
        long as we are still running.

        This method exits when one of the following occurs:

        * The callable function completes.
        * This server stops.
        * The lock (if supplied) is invalidated due to connection loss.
        * A non-retryable ZK exception is raised.

        Pass None as the lock parameter if the lock issue is not
        relevant.

        """
        def interrupt():
            if ((not self.isComponentRunning()) or
                (lock and not lock.is_still_valid())):
                return True
            return False

        kazoo_retry = KazooRetry(max_tries=-1, interrupt=interrupt,
                                 delay=5, backoff=0, ignore_expire=False)
        try:
            return kazoo_retry(func, *args, **kw)
        except InterruptedError:
            pass


class MergeServer(BaseMergeServer):
    log = logging.getLogger("zuul.MergeServer")

    def __init__(
        self,
        config: ConfigParser,
        connections,
    ):
        super().__init__(config, 'merger', connections)
        self.hostname = socket.getfqdn()
        self.component_info = MergerComponent(
            self.zk_client, self.hostname, version=get_version_string())
        self.component_info.register()
        COMPONENT_REGISTRY.create(self.zk_client)

        self.monitoring_server = MonitoringServer(self.config, 'merger',
                                                  self.component_info)
        self.monitoring_server.start()

        self.command_map = {
            commandsocket.StopCommand.name: self.stop,
            # Stop for the mergers is always graceful. We add this alias
            # to make it clearer to users that they can gracefully stop.
            commandsocket.GracefulCommand.name: self.stop,
            commandsocket.PauseCommand.name: self.pause,
            commandsocket.UnPauseCommand.name: self.unpause,
        }
        command_socket = get_default(
            self.config, 'merger', 'command_socket',
            '/var/lib/zuul/merger.socket')
        self.command_socket = commandsocket.CommandSocket(command_socket)

        self._command_running = False

    def start(self):
        super().start()
        self._command_running = True
        self.log.debug("Starting command processor")
        self.command_socket.start()
        self.command_thread = threading.Thread(
            target=self.runCommand, name='command')
        self.command_thread.daemon = True
        self.command_thread.start()
        self.component_info.state = self.component_info.RUNNING

    def stop(self):
        self.log.debug("Stopping merger")
        self.component_info.state = self.component_info.STOPPED
        super().stop()
        self._command_running = False
        self.command_socket.stop()
        self.monitoring_server.stop()
        self.log.debug("Stopped merger")

    def join(self):
        self.log.debug("Joining merger")
        super().join()
        self.monitoring_server.join()
        self.log.debug("Joined merger")

    def pause(self):
        self.log.debug('Pausing')
        self.component_info.state = self.component_info.PAUSED
        super().pause()

    def unpause(self):
        self.log.debug('Resuming')
        super().unpause()
        self.component_info.state = self.component_info.RUNNING

    def runCommand(self):
        while self._command_running:
            try:
                command, args = self.command_socket.get()
                if command != '_stop':
                    self.command_map[command](*args)
            except Exception:
                self.log.exception("Exception while processing command")



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This adds some helpers that are useful for mutating hyphenated YAML
# structures into underscored python dicts.

import voluptuous as vs


UNDEFINED = object()


def assemble(*schemas, doc=None):
    """Merge any number of voluptuous schemas into a single schema.  The
    input schemas must all be dictionary-based.

    """
    ret = vs.Schema({})
    old_doc = None
    for x in schemas:
        ret = ret.extend(x.schema)
        if schema_doc := getattr(x, 'doc', None):
            old_doc = schema_doc
    doc = doc or old_doc
    if doc:
        ret.doc = doc
    return ret


def discriminate(comparator):
    """Chose among multiple schemas in a Union using a supplied comparator
    function.  This does some extra work to find the interesting part
    of nested schemas.  It follows nested schemas down and picks the
    first one of each.  This makes it suitable for use with All
    combined with RequiredExclusive.  This is intended to be passed to
    the discriminant argument of Union.

    The comparator is called with the following arguments:

    :param object val: The input data to be validated.
    :param Schema s: The first schema with values likely to be comparable.

    """
    def disc(val, alt):
        ret = []
        for a in alt:
            s = a
            while hasattr(s, 'validators'):
                s = s.validators[0]
            while hasattr(s, 'schema'):
                s = s.schema
            if comparator(val, s):
                ret.append(a)
        return ret
    return disc


class RequiredExclusive:
    """A validator to require that at least one of a set of alternates is
    present.  Use this in combination with Exclusive and All (since
    Exclusive doesn't imply Required).  Example:

    All(
        {Exclusive(...), Exclusive(...)},
        RequiredExclusive(...)
    )
    """
    def __init__(self, *attributes, msg=None, doc=None):
        self.schema = vs.Schema({
            # Use the mutated names for the requirements
            vs.Required(
                vs.Any(*attributes),
                msg=msg,
            ): object,
            object: object,
        })
        self.doc = doc

    def __call__(self, v):
        return self.schema(v)


class Required(vs.Required):
    """Require an attribute and mutate its name

    This mutates the name of the attribute to lowercase it and replace
    hyphens with underscares.  This enables the output of the
    validator to be used directly in setting python object attributes.

    """
    def __init__(self, schema, default=UNDEFINED, output=None, doc=None):
        if not isinstance(schema, str):
            raise Exception("Only strings are supported")
        super().__init__(schema, default=default)
        if output is None:
            output = str(schema).replace('-', '_').lower()
        self.output = output
        self.doc = doc

    def __call__(self, data):
        # Superclass ensures that data==schema
        super().__call__(data)
        # Return our mutated form
        return self.output


class Optional(vs.Optional):
    """Mark an attribute optional and mutate its name

    This mutates the name of the attribute to lowercase it and replace
    hyphens with underscares.  This enables the output of the
    validator to be used directly in setting python object attributes.

    This works in conjuction with Nullable.
    """
    def __init__(self, schema, default=UNDEFINED, output=None, doc=None):
        if not isinstance(schema, str):
            raise Exception("Only strings are supported")
        super().__init__(schema, default=default)
        if output is None:
            output = str(schema).replace('-', '_').lower()
        self.output = output
        self.doc = doc

    def __call__(self, data):
        # Superclass ensures that data==schema
        super().__call__(data)
        # Return our mutated form
        return self.output


class Nullable:
    """Set the output value to None when no input is supplied.

    When used with Optional, if no input value is supplied, this will
    set the output to None without also accepting None as an input
    value.

    """
    def __init__(self, schema):
        self.schema = vs.Schema(schema)

    def __call__(self, v):
        if v is UNDEFINED:
            return None
        return self.schema(v)


class AsList:
    """Accept either the specified value or a list of it.

    Always return a list.
    """
    def __init__(self, schema):
        self.schema = vs.Any([schema], schema)

    def __call__(self, v):
        val = self.schema(v)
        if not val:
            return []
        if isinstance(val, list):
            return val
        return [val]


class Constant:
    """A string constant with a documentation attribute

    Use this instead of a bare string when you want to supply
    documentation.

    """
    def __init__(self, schema, doc=None):
        self.schema = vs.Schema(schema)
        self.schema.doc = doc

    def __call__(self, v):
        return self.schema(v)

    def __eq__(self, other):
        return str(self.schema) == other


class Exclusive(vs.Exclusive):
    """Subclass of Exclusive that supports documentation
    """
    def __init__(self, schema, group_of_exclusion, msg=None,
                 description=None, doc=None):
        super().__init__(schema, group_of_exclusion, msg, description)
        self.doc = schema.doc



# Copyright 2017 Red Hat, Inc.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.model import TriggerEvent
from zuul.model import EventFilter

EMPTY_GIT_REF = '0' * 40  # git sha of all zeros, used during creates/deletes


class GitTriggerEvent(TriggerEvent):
    """Incoming event from an external system."""

    def __repr__(self):
        ret = '<GitTriggerEvent %s %s' % (self.type,
                                          self.project_name)

        if self.branch:
            ret += " branch:%s" % self.branch
        if self.ref:
            ret += " ref:%s" % self.ref
        ret += " oldrev:%s" % self.oldrev
        ret += " newrev:%s" % self.newrev
        ret += '>'

        return ret


class GitEventFilter(EventFilter):
    def __init__(self, connection_name, trigger, types=None, refs=None,
                 ignore_deletes=True, debug=None):

        super().__init__(connection_name, trigger, debug)

        refs = refs if refs is not None else []
        self._refs = [x.pattern for x in refs]
        self.refs = refs

        self.types = types if types is not None else []
        self.ignore_deletes = ignore_deletes

    def __repr__(self):
        ret = '<GitEventFilter'
        ret += ' connection: %s' % self.connection_name

        if self.types:
            ret += ' types: %s' % ', '.join(self.types)
        if self._refs:
            ret += ' refs: %s' % ', '.join(self._refs)
        if self.ignore_deletes:
            ret += ' ignore_deletes: %s' % self.ignore_deletes
        ret += '>'

        return ret

    def matches(self, event, change):
        if not super().matches(event, change):
            return False

        # event types are ORed
        matches_type = False
        for etype in self.types:
            if etype == event.type:
                matches_type = True
        if self.types and not matches_type:
            return False

        # refs are ORed
        matches_ref = False
        if event.ref is not None:
            for ref in self.refs:
                if ref.match(event.ref):
                    matches_ref = True
        if self.refs and not matches_ref:
            return False
        if self.ignore_deletes and event.newrev == EMPTY_GIT_REF:
            # If the updated ref has an empty git sha (all 0s),
            # then the ref is being deleted
            return False

        return True



# Copyright (c) 2017 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from tests.base import ZuulGithubAppTestCase, ZuulTestCase, simple_layout


class TestGithubRequirements(ZuulTestCase):
    """Test pipeline and trigger requirements"""
    config_file = 'zuul-github-driver.conf'
    scheduler_count = 1

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_require_status(self):
        "Test pipeline requirement: status"
        project = 'org/project1'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No status from zuul so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # An error status should not cause it to be enqueued
        self.fake_github.setCommitStatus(project, A.head_sha, 'error',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A success status goes in
        self.fake_github.setCommitStatus(project, A.head_sha, 'success',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project1-pipeline')

        # Trigger regex matched status
        self.fake_github.emitEvent(A.getCommentAddedEvent('test regex'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, 'project1-pipeline')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_trigger_require_status(self):
        "Test trigger requirement: status"
        project = 'org/project1'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('trigger me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No status from zuul so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # An error status should not cause it to be enqueued
        self.fake_github.setCommitStatus(project, A.head_sha, 'error',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A success status goes in
        self.fake_github.setCommitStatus(project, A.head_sha, 'success',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project1-pipeline')

        self.fake_github.emitEvent(A.getCommentAddedEvent('trigger regex'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, 'project1-pipeline')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_trigger_on_status(self):
        "Test trigger on: status"
        project = 'org/project2'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')

        # Create second PR which contains the head of A in its history. Zuul
        # should not get disturbed by the existence of this one.
        self.fake_github.openFakePullRequest(
            project, 'master', 'A', base_sha=A.head_sha)

        # An error status should not cause it to be enqueued
        self.fake_github.setCommitStatus(project, A.head_sha, 'error',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(A.getCommitStatusEvent('tenant-one/check',
                                                          state='error'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A success status from unknown user should not cause it to be
        # enqueued
        self.fake_github.setCommitStatus(project, A.head_sha, 'success',
                                         context='tenant-one/check',
                                         user='foo')
        self.fake_github.emitEvent(A.getCommitStatusEvent('tenant-one/check',
                                                          state='success',
                                                          user='foo'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A success status from zuul goes in
        self.fake_github.setCommitStatus(project, A.head_sha, 'success',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(A.getCommitStatusEvent('tenant-one/check'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project2-trigger')

        # An error status for a different context should not cause it to be
        # enqueued
        self.fake_github.setCommitStatus(project, A.head_sha, 'error',
                                         context='tenant-one/gate')
        self.fake_github.emitEvent(A.getCommitStatusEvent('tenant-one/gate',
                                                          state='error'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

        # A success status with a regex match goes in
        self.fake_github.emitEvent(A.getCommitStatusEvent('cooltest',
                                                          user='other-ci'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, 'project2-trigger')

    @simple_layout("layouts/requirements-github.yaml", driver="github")
    def test_trigger_on_check_run(self):
        """Test trigger on: check_run"""
        project = "org/project15"
        A = self.fake_github.openFakePullRequest(project, "master", "A")

        # A check_run request with a different name should not cause it to be
        # enqueued.
        self.fake_github.emitEvent(
            A.getCheckRunRequestedEvent("tenant-one/different-check")
        )
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A check_run request with the correct name, but for a different app
        # should not cause it to be enqueued.
        self.fake_github.emitEvent(
            A.getCheckRunRequestedEvent("tenant-one/check", app="other-ci")
        )
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A check_run request with the correct name for the correct app should
        # cause it to be enqueued.
        self.fake_github.emitEvent(
            A.getCheckRunRequestedEvent("tenant-one/check"))

        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, "project15-check-run")

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_require_review_username(self):
        "Test pipeline requirement: review username"

        A = self.fake_github.openFakePullRequest('org/project3', 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No approval from derp so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # Add an approved review from derp
        A.addReview('derp', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project3-reviewusername')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_require_review_state(self):
        "Test pipeline requirement: review state"

        A = self.fake_github.openFakePullRequest('org/project4', 'master', 'A')
        # Add derp to writers
        A.writers.extend(('derp', 'werp'))
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No positive review from derp so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # A negative review from derp should not cause it to be enqueued
        A.addReview('derp', 'CHANGES_REQUESTED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A negative review from werp should not cause it to be enqueued
        A.addReview('werp', 'CHANGES_REQUESTED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive from nobody should not cause it to be enqueued
        A.addReview('nobody', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive review from derp should still be blocked by the
        # negative review from werp
        A.addReview('derp', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive review from werp should cause it to be enqueued
        A.addReview('werp', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project4-reviewreq')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_require_review_user_state(self):
        "Test pipeline requirement: review state from user"

        A = self.fake_github.openFakePullRequest('org/project5', 'master', 'A')
        # Add derp and herp to writers
        A.writers.extend(('derp', 'herp'))
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No positive review from derp so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # A negative review from derp should not cause it to be enqueued
        A.addReview('derp', 'CHANGES_REQUESTED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive from nobody should not cause it to be enqueued
        A.addReview('nobody', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive review from herp (a writer) should not cause it to be
        # enqueued
        A.addReview('herp', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive review from derp should cause it to be enqueued
        A.addReview('derp', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project5-reviewuserstate')

# TODO: Implement reject on approval username/state

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_require_review_latest_user_state(self):
        "Test pipeline requirement: review state from user"

        A = self.fake_github.openFakePullRequest('org/project5', 'master', 'A')
        # Add derp and herp to writers
        A.writers.extend(('derp', 'herp'))
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No positive review from derp so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # The first negative review from derp should not cause it to be
        # enqueued
        A.addReview('derp', 'CHANGES_REQUESTED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive review from derp should cause it to be enqueued
        A.addReview('derp', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project5-reviewuserstate')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_require_review_write_perms(self):
        "Test pipeline requirement: review from user with write"

        A = self.fake_github.openFakePullRequest('org/project4', 'master', 'A')
        # Add herp to admins
        A.admins.append('herp')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No positive review from derp so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # The first review is from a reader, and thus should not be enqueued
        A.addReview('derp', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive review from herp should cause it to be enqueued
        A.addReview('herp', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project4-reviewreq')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_require_review_comment_masked(self):
        "Test pipeline requirement: review comments on top of votes"

        A = self.fake_github.openFakePullRequest('org/project5', 'master', 'A')
        # Add derp to writers
        A.writers.append('derp')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No positive review from derp so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # The first negative review from derp should not cause it to be
        # enqueued
        A.addReview('derp', 'CHANGES_REQUESTED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive review is required, so provide it
        A.addReview('derp', 'APPROVED')

        # Add a comment review on top to make sure we can still enqueue
        A.addReview('derp', 'COMMENTED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project5-reviewuserstate')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_require_review_newer_than(self):

        A = self.fake_github.openFakePullRequest('org/project6', 'master', 'A')
        # Add derp and herp to writers
        A.writers.extend(('derp', 'herp'))
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No positive review from derp so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # Add a too-old positive review, should not be enqueued
        submitted_at = time.time() - 72 * 60 * 60
        A.addReview('derp', 'APPROVED',
                    submitted_at)
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Add a recent positive review
        submitted_at = time.time() - 12 * 60 * 60
        A.addReview('derp', 'APPROVED', submitted_at)
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project6-newerthan')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_require_review_older_than(self):

        A = self.fake_github.openFakePullRequest('org/project7', 'master', 'A')
        # Add derp and herp to writers
        A.writers.extend(('derp', 'herp'))
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No positive review from derp so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # Add a too-new positive, should not be enqueued
        submitted_at = time.time() - 12 * 60 * 60
        A.addReview('derp', 'APPROVED',
                    submitted_at)
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Add an old enough positive, should enqueue
        submitted_at = time.time() - 72 * 60 * 60
        A.addReview('herp', 'APPROVED', submitted_at)
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project7-olderthan')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_require_open(self):

        A = self.fake_github.openFakePullRequest('org/project8', 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()

        # PR is open, we should have enqueued
        self.assertEqual(len(self.history), 1)

        # close the PR and try again
        A.state = 'closed'
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # PR is closed, should not trigger
        self.assertEqual(len(self.history), 1)

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_reject_open(self):

        A = self.fake_github.openFakePullRequest('org/project13', 'master',
                                                 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()

        # PR is open, we should not have enqueued
        self.assertEqual(len(self.history), 0)

        # close the PR and try again
        A.state = 'closed'
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # PR is closed, should trigger
        self.assertEqual(len(self.history), 1)

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_require_current(self):

        A = self.fake_github.openFakePullRequest('org/project9', 'master',
                                                 'A')
        # A sync event that we will keep submitting to trigger
        sync = A.getPullRequestSynchronizeEvent()
        self.fake_github.emitEvent(sync)
        self.waitUntilSettled()

        # PR head is current should enqueue
        self.assertEqual(len(self.history), 1)

        # Add a commit to the PR, re-issue the original comment event
        A.addCommit()
        self.fake_github.emitEvent(sync)
        self.waitUntilSettled()
        # Event hash is not current, should not trigger
        self.assertEqual(len(self.history), 1)

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_reject_current(self):

        A = self.fake_github.openFakePullRequest('org/project14', 'master',
                                                 'A')
        # A sync event that we will keep submitting to trigger
        sync = A.getPullRequestSynchronizeEvent()
        self.fake_github.emitEvent(sync)
        self.waitUntilSettled()

        # PR head is current, should not enqueue
        self.assertEqual(len(self.history), 0)

        # Add a commit to the PR, re-issue the original comment event
        A.addCommit()
        self.fake_github.emitEvent(sync)
        self.waitUntilSettled()
        # Event hash is not current, should trigger
        self.assertEqual(len(self.history), 1)

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_require_draft(self):

        A = self.fake_github.openFakePullRequest('org/project17', 'master',
                                                 'A', draft=True)
        # A sync event that we will keep submitting to trigger
        sync = A.getPullRequestSynchronizeEvent()
        self.fake_github.emitEvent(sync)
        self.waitUntilSettled()

        # PR is a draft, should enqueue
        self.assertEqual(len(self.history), 1)

        # Make the PR not a draft
        A.draft = False
        self.fake_github.emitEvent(sync)
        self.waitUntilSettled()
        # PR is not a draft, should not enqueue
        self.assertEqual(len(self.history), 1)

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_reject_draft(self):

        A = self.fake_github.openFakePullRequest('org/project18', 'master',
                                                 'A', draft=True)
        # A sync event that we will keep submitting to trigger
        sync = A.getPullRequestSynchronizeEvent()
        self.fake_github.emitEvent(sync)
        self.waitUntilSettled()

        # PR is a draft, should not enqueue
        self.assertEqual(len(self.history), 0)

        # Make the PR not a draft
        A.draft = False
        self.fake_github.emitEvent(sync)
        self.waitUntilSettled()
        # PR is not a draft, should enqueue
        self.assertEqual(len(self.history), 1)

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_require_label(self):
        "Test pipeline requirement: label"
        A = self.fake_github.openFakePullRequest('org/project10', 'master',
                                                 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No label so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # A derp label should not cause it to be enqueued
        A.addLabel('derp')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # An approved label goes in
        A.addLabel('approved')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project10-label')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_reject_label(self):
        "Test pipeline reject: label"
        A = self.fake_github.openFakePullRequest('org/project11', 'master',
                                                 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # No label so should not be enqueued
        self.assertEqual(len(self.history), 0)

        # A do-not-merge label should not cause it to be enqueued
        A.addLabel('do-not-merge')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # An approved label should still not enqueue due to d-n-m
        A.addLabel('approved')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Remove do-not-merge should enqueue
        A.removeLabel('do-not-merge')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project11-label')

    @simple_layout('layouts/requirements-github.yaml', driver='github')
    def test_pipeline_reject_status(self):
        "Test pipeline reject: status"
        project = 'org/project12'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')

        # Set rejected error status
        self.fake_github.setCommitStatus(project, A.head_sha, 'error',
                                         context='tenant-one/check')

        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent('test me')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        # Status should cause it to be rejected
        self.assertEqual(len(self.history), 0)

        # Test that also the regex matched pipeline doesn't trigger
        self.fake_github.emitEvent(A.getCommentAddedEvent('test regex'))
        self.waitUntilSettled()
        # Status should cause it to be rejected
        self.assertEqual(len(self.history), 0)

        self.fake_github.setCommitStatus(project, A.head_sha, 'success',
                                         context='tenant-one/check')
        # Now that status is not error, it should be enqueued
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, 'project12-status')

        # Test that also the regex matched pipeline triggers now
        self.fake_github.emitEvent(A.getCommentAddedEvent('test regex'))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, 'project12-status')


class TestGithubAppRequirements(ZuulGithubAppTestCase):
    """Test pipeline and trigger requirements with app authentication"""
    config_file = 'zuul-github-driver.conf'
    scheduler_count = 1

    @simple_layout("layouts/requirements-github.yaml", driver="github")
    def test_pipeline_require_check_run(self):
        "Test pipeline requirement: status (reported via a check run)"
        project = "org/project16"
        github = self.fake_github.getGithubClient()
        repo = github.repo_from_project(project)

        A = self.fake_github.openFakePullRequest(project, "master", "A")
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent("trigger me")
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()

        # No status from zuul, so nothing should be enqueued
        self.assertEqual(len(self.history), 0)

        # An error check run should also not cause it to be enqueued
        repo.create_check_run(
            A.head_sha,
            "tenant-one/check",
            conclusion="failure",
            app="check-run",
        )
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A success check run goes in, ready to be enqueued
        repo.create_check_run(
            A.head_sha,
            "tenant-one/check",
            conclusion="success",
            app="check-run",
        )
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)


class TestGithubTriggerRequirements(ZuulTestCase):
    """Test pipeline and trigger requirements"""
    config_file = 'zuul-github-driver.conf'
    scheduler_count = 1

    @simple_layout('layouts/github-trigger-requirements.yaml', driver='github')
    def test_require_status(self):
        # Test trigger require-status
        jobname = 'require-status'
        project = 'org/project'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent(f'test {jobname}')

        # No status from zuul so should not be enqueued
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # An error status should not cause it to be enqueued
        self.fake_github.setCommitStatus(project, A.head_sha, 'error',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A success status goes in
        self.fake_github.setCommitStatus(project, A.head_sha, 'success',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/github-trigger-requirements.yaml', driver='github')
    def test_reject_status(self):
        # Test trigger reject-status
        jobname = 'reject-status'
        project = 'org/project'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent(f'test {jobname}')

        # No status from zuul so should be enqueued
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

        # A failure status should not cause it to be enqueued
        self.fake_github.setCommitStatus(project, A.head_sha, 'failure',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

        # A success status goes in
        self.fake_github.setCommitStatus(project, A.head_sha, 'success',
                                         context='tenant-one/check')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, jobname)

    @simple_layout('layouts/github-trigger-requirements.yaml', driver='github')
    def test_require_review(self):
        # Test trigger require-review
        jobname = 'require-review'
        project = 'org/project'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')
        A.writers.extend(('maintainer',))
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent(f'test {jobname}')

        # No review so should not be enqueued
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # An changes requested review should not cause it to be enqueued
        A.addReview('maintainer', 'CHANGES_REQUESTED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A positive review goes in
        A.addReview('maintainer', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/github-trigger-requirements.yaml', driver='github')
    def test_reject_review(self):
        # Test trigger reject-review
        jobname = 'reject-review'
        project = 'org/project'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')
        A.writers.extend(('maintainer',))
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent(f'test {jobname}')

        # No review so should be enqueued
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

        # An changes requested review should not cause it to be enqueued
        A.addReview('maintainer', 'CHANGES_REQUESTED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

        # A positive review goes in
        A.addReview('maintainer', 'APPROVED')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, jobname)

    @simple_layout('layouts/github-trigger-requirements.yaml', driver='github')
    def test_require_label(self):
        # Test trigger require-label
        jobname = 'require-label'
        project = 'org/project'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent(f'test {jobname}')

        # No label so should not be enqueued
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # A random should not cause it to be enqueued
        A.addLabel('foobar')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # An approved label goes in
        A.addLabel('approved')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

    @simple_layout('layouts/github-trigger-requirements.yaml', driver='github')
    def test_reject_label(self):
        # Test trigger reject-label
        jobname = 'reject-label'
        project = 'org/project'
        A = self.fake_github.openFakePullRequest(project, 'master', 'A')
        # A comment event that we will keep submitting to trigger
        comment = A.getCommentAddedEvent(f'test {jobname}')

        # No label so should be enqueued
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(self.history[0].name, jobname)

        # A rejected label should not cause it to be enqueued
        A.addLabel('rejected')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)

        # Any other label, it goes in
        A.removeLabel('rejected')
        A.addLabel('okay')
        self.fake_github.emitEvent(comment)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(self.history[1].name, jobname)



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import os
import subprocess
import tempfile

from zuul.lib import encryption

from tests.base import BaseTestCase


class TestEncryption(BaseTestCase):

    def setUp(self):
        super(TestEncryption, self).setUp()
        self.private, self.public = encryption.generate_rsa_keypair()
        # Because we set delete to False when using NamedTemporaryFile below
        # we need to stick our usage of temporary files in the NestedTempfile
        # fixture ensuring everything gets cleaned up when it is done.
        self.useFixture(fixtures.NestedTempfile())

    def test_serialization(self):
        "Verify key serialization"
        pem_private = encryption.serialize_rsa_private_key(self.private)
        private2, public2 = encryption.deserialize_rsa_keypair(pem_private)

        # cryptography public / private key objects don't implement
        # equality testing, so we make sure they have the same numbers.
        self.assertEqual(self.private.private_numbers(),
                         private2.private_numbers())
        self.assertEqual(self.public.public_numbers(),
                         public2.public_numbers())

    def test_pkcs1_oaep(self):
        "Verify encryption and decryption"
        orig_plaintext = b"some text to encrypt"
        ciphertext = encryption.encrypt_pkcs1_oaep(orig_plaintext, self.public)
        plaintext = encryption.decrypt_pkcs1_oaep(ciphertext, self.private)
        self.assertEqual(orig_plaintext, plaintext)

    def test_openssl_pkcs1_oaep(self):
        "Verify that we can decrypt something encrypted with OpenSSL"
        orig_plaintext = b"some text to encrypt"
        pem_public = encryption.serialize_rsa_public_key(self.public)
        public_file = tempfile.NamedTemporaryFile(delete=False)
        try:
            public_file.write(pem_public)
            public_file.close()

            p = subprocess.Popen(['openssl', 'rsautl', '-encrypt',
                                  '-oaep', '-pubin', '-inkey',
                                  public_file.name],
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE)
            (stdout, stderr) = p.communicate(orig_plaintext)
            ciphertext = stdout
        finally:
            os.unlink(public_file.name)

        plaintext = encryption.decrypt_pkcs1_oaep(ciphertext, self.private)
        self.assertEqual(orig_plaintext, plaintext)



#!/usr/bin/env python
# Copyright (C) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This script updates the Zuul v3 Storyboard.  It uses a .boartty.yaml
# file to get credential information.

import requests
import boartty.config
import boartty.sync
import logging  # noqa
from pprint import pprint as p  # noqa


class App(object):
    pass


def get_tasks(sync):
    task_list = []
    for story in sync.get('/v1/stories?tags=zuulv3'):
        print("Story %s: %s" % (story['id'], story['title']))
        for task in sync.get('/v1/stories/%s/tasks' % (story['id'])):
            print("  %s" % (task['title'],))
            task_list.append(task)
    return task_list


def task_in_lane(task, lane):
    for item in lane['worklist']['items']:
        if 'task' in item and item['task']['id'] == task['id']:
            return True
    return False


def add_task(sync, task, lane):
    print("Add task %s to %s" % (task['id'], lane['worklist']['id']))
    r = sync.post('v1/worklists/%s/items/' % lane['worklist']['id'],
                  dict(item_id=task['id'],
                       item_type='task',
                       list_position=0))
    print(r)


def remove_task(sync, task, lane):
    print("Remove task %s from %s" % (task['id'], lane['worklist']['id']))
    for item in lane['worklist']['items']:
        if 'task' in item and item['task']['id'] == task['id']:
            r = sync.delete('v1/worklists/%s/items/' % lane['worklist']['id'],
                            dict(item_id=item['id']))
            print(r)


MAP = {
    'todo': ['New', 'Backlog', 'Todo'],
    'inprogress': ['In Progress', 'Blocked'],
    'review': ['In Progress', 'Blocked'],
    'merged': None,
    'invalid': None,
}


def main():
    requests.packages.urllib3.disable_warnings()
    # logging.basicConfig(level=logging.DEBUG)
    app = App()
    app.config = boartty.config.Config('openstack')
    sync = boartty.sync.Sync(app, False)
    board = sync.get('v1/boards/41')
    tasks = get_tasks(sync)

    lanes = dict()
    for lane in board['lanes']:
        lanes[lane['worklist']['title']] = lane

    for task in tasks:
        ok_lanes = MAP[task['status']]
        task_found = False
        for lane_name, lane in lanes.items():
            if task_in_lane(task, lane):
                if ok_lanes and lane_name in ok_lanes:
                    task_found = True
                else:
                    remove_task(sync, task, lane)
        if ok_lanes and not task_found:
            add_task(sync, task, lanes[ok_lanes[0]])


if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""zuul_ref_indexes

Revision ID: 3c1488fb137e
Revises: d0a269345a22
Create Date: 2023-11-27 17:15:00.812075

"""

# revision identifiers, used by Alembic.
revision = '3c1488fb137e'
down_revision = 'd0a269345a22'
branch_labels = None
depends_on = None

from alembic import op

REF_TABLE = 'zuul_ref'


def upgrade(table_prefix=''):
    prefixed_ref = table_prefix + REF_TABLE

    op.create_index(f'{prefixed_ref}_project_idx',
                    prefixed_ref,
                    ['project'])
    op.create_index(f'{prefixed_ref}_ref_idx',
                    prefixed_ref,
                    ['ref'])
    op.create_index(f'{prefixed_ref}_branch_idx',
                    prefixed_ref,
                    ['branch'])
    # Change is already indexed
    op.create_index(f'{prefixed_ref}_patchset_idx',
                    prefixed_ref,
                    ['patchset'])
    op.create_index(f'{prefixed_ref}_oldrev_idx',
                    prefixed_ref,
                    ['oldrev'])
    op.create_index(f'{prefixed_ref}_newrev_idx',
                    prefixed_ref,
                    ['newrev'])

    # Drop the project_change index and rely on separate indexes that
    # may be merged by the dbms.
    op.drop_index(f'{prefixed_ref}_project_change_idx', prefixed_ref)


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import re
import yaml

from zuul.web import ZuulWeb, ZuulWebAPI


PARAM_RE = re.compile(':param (.*?) (.*?): (.*)')

spec = {
    'info': {
        'title': 'Zuul REST API',
        'version': 'v1',
        'description': 'Incomplete (work in progress) list of the endpoints.',
    },
    'openapi': '3.0.0',
    'tags': [
        {'name': 'tenant'},
    ],
    'paths': {},
}


def parse_docstring(doc):
    # This separates the overall method summary (to be used as the
    # endpoint summary in the openapi spec) from parameter types and
    # descriptions.
    summary = []
    params = {}
    pname = None
    ptype = None
    pbuf = []
    for line in doc.split('\n'):
        line = line.strip()
        if not line:
            continue
        m = PARAM_RE.match(line)
        if m:
            if pname:
                params[pname] = {
                    'type': ptype,
                    'desc': ' '.join(pbuf),
                }
            ptype, pname, pdesc = m.groups()
            if ptype == 'str':
                ptype = 'string'
            pbuf = [pdesc.strip()]
            continue
        if pname:
            pbuf.append(line.strip())
        else:
            summary.append(line.strip())
    if pname:
        params[pname] = {
            'type': ptype,
            'desc': ' '.join(pbuf),
        }
    return ' '.join(summary), params


def generate_spec():
    api = ZuulWebAPI
    route_map = ZuulWeb.generateRouteMap(api, None, True)
    # Iterate over each route
    for r in route_map.mapper.matchlist:
        # Some of our routes have globs in the variable names; remove
        # those so that "{project:.*}" becomes "{project}" to match
        # the function arguments.
        routepath = r.routepath.replace(':.*', '')
        # action is the ZuulWebAPI method name
        action = r.defaults['action']
        # handler is the ZuulWebAPI method itself
        handler = getattr(api, action, None)
        if handler is None:
            continue
        # This is our output; initialize it to an empty dict, or if
        # we're handling another instance of a previous route (for
        # example, different GET and POST methods), start with that.
        routespec = spec['paths'].setdefault(routepath, {})
        # Parse the docstring if available
        doc = handler.__doc__
        if doc:
            summary, doc_params = parse_docstring(doc)
        else:
            summary = ''
            doc_params = {}
        if r.conditions and r.conditions.get('method'):
            # If this route specifies methods, use that
            methods = r.conditions['method']
        else:
            # Otherwise assume this method only handles GET
            methods = ['GET']
        for method in methods:
            if method == 'OPTIONS':
                continue
            # The @openapi decorators set this attribute; initialize
            # the output dictionary for this route-method to that
            # value or the empty dict.
            default_methodspec = {
                'responses': {
                    '200': {
                        'description': 'Response not yet documented',
                    }
                }
            }
            methodspec = getattr(handler, '__openapi__', default_methodspec)
            routespec[method.lower()] = methodspec
            methodspec['summary'] = summary
            methodspec['operationId'] = action
            methodspec['tags'] = ['tenant']
            methodspec['parameters'] = []

            # All inputs should be in the method signature, so iterate
            # over that.
            for handler_param in inspect.signature(
                    handler).parameters.values():
                paramspec = {}
                # See if this function argument appears in the route
                # path, if so, it's a "path" parameter, otherwise it's
                # a "query" param.
                in_path = '{' + handler_param.name + '}' in routepath
                required = False
                if handler_param.default is handler_param.empty:
                    # No default value; it's either in the path or we
                    # don't care about it.
                    if not in_path:
                        continue
                    required = True
                paramspec = {
                    'name': handler_param.name,
                    'in': 'path' if in_path else 'query',
                }
                # Merge in information from the docstring if available.
                doc_param = doc_params.get(handler_param.name)
                if doc_param:
                    paramspec['description'] = doc_param['desc']
                    paramspec['schema'] = {'type': doc_param['type']}
                else:
                    paramspec['schema'] = {'type': 'string'}
                if required:
                    paramspec['required'] = required

                methodspec['parameters'].append(paramspec)
    return spec


def main():
    with open('web/public/openapi.yaml', 'w') as f:
        f.write(yaml.safe_dump(generate_spec(),
                               default_flow_style=False))


if __name__ == '__main__':
    main()
